from pyg_base._types import is_int, is_float, is_str, is_date, is_bool
from pyg_base._dates import dt
from pyg_base._loop import loop
from enum import Enum

@loop(list, tuple)
def _as_primitive(value):
    if is_bool(value):
        return True if value else False
    elif is_int(value):
        return int(value)
    elif is_float(value):
        return float(value)
    elif is_date(value):
        return dt(value)
    elif value is None or is_str(value):
        return value
    elif isinstance(value, Enum):
        return _as_primitive(value.value)
    else:
        return value

def as_primitive(value):
    return _as_primitive(value)
    



import os
import json
CFG = os.environ.get('PYG_CFG')

def mkdir(path):
    """
    sames as os.mkdir but 
    1) allows for subdir i.e.:
    2) will not attempt to create if path exits, hence can run safely


    :Example:
    ---------
    >>> with pytest.raises(Exception):       
    >>>     os.mkdir('c:/temp/some/subdir/that/we/want')

    >>> print('but', mkdir('c:/temp/some/subdir/that/we/want'), 'now exists')
    
    """
    if not os.path.isdir(path):
        p = path.replace('\\', '/')
        paths = p.split('/')
        if '.' in paths[-1]:
            paths = paths[:-1]
        for i in range(2, len(paths)+1):
            d = '/'.join(paths[:i])
            if not os.path.isdir(d):
                os.mkdir(d)
    return path

CACHE = {}

def get_cache(*args):
    res = CACHE
    for arg in args:
        if arg not in res:
            res[arg] = {}
        res = res[arg]
    return res

def cfg_read():
    cfg = CACHE.get('CFG', {})
    if isinstance(CFG, str):
        for path in CFG.split(','):
            if os.path.isfile(path):
                with open(path, 'r') as f:
                    cfg.update(json.load(f))
                    CACHE['CFG'] = cfg
    return cfg

cfg_read.__doc__ = f'reads the config from get_cache("CFG") and updates it from files in {CFG}'


def cfg_write(cfg):
    CACHE['CFG'] = cfg
    if CFG is not None:
        for path in CFG.split(','):
            try:
                with open(mkdir(path), 'w') as f:
                    json.dump(cfg, f)
                return
            except Exception:
                pass
   
cfg_write.__doc__ = f'writes the config file provided to {CFG}' 
    


### we create an interface for read/write of pandas which is bitemporal
# the data
from pyg_base._types import is_df, is_date, is_series, is_int
from pyg_base._dates import dt, dt_bump, is_bump
from pyg_base._as_list import as_list
from pyg_base._loop import loop
from pyg_base._sort import sort

import pandas as pd
import numpy as np
from functools import partial 

_series = '_is_series'
_columns = '_column_names'
_asof = '_asof'

def is_bi(df):
    return is_df(df) and _asof in df.columns

_nth = lambda v, n: v.iloc[min(n, len(v)-1)] if n>=0 else v.iloc[max(n, -len(v))]
    
def _as_what(what):
    if is_int(what):
        return partial(_nth, n = what)
    else:
        return what
    
    
def bi_read(df, asof = None, what = -1):
    """
    Parameters
    ----------
    df : a timeseries
        an item we can read bitemporally    
        
    asof: datetime
        read the values as if date is asof
        
    what: int, str or function
        how to select a value from multiple values associated with same index
        'first'  or 'last' would be the first non-nan values.
        0 : first actual value that was observed, even if nan
        -1: last actual value observed, even if nan
        -2,-3...    : one or two before last if available, else first
        1,2...    : 2nd or 3rd releases if available, else last
    if what is 'all' then the raw data is returned
    
    Example:
    --------
        df = pd.DataFrame(dict(a = [1,2,3,4,5,6], 
            _asof = drange(-2) + drange(-1) + drange(0)), 
            index = [dt(-2)] * 3 + [dt(-1)] * 2 + [dt(0)])
    
    """
    if not is_bi(df) or what == 'all':
        return df
    if is_date(asof):
        df = df[df[_asof]<=asof]
    if is_bi(asof):
        df = df[df[_asof] <= asof.reindex(df.index)[_asof]]
    index_name = df.index.name
    if len(df):        
        if index_name is None:
            df.index.name = 'index'
        gb = df.sort_values(_asof).groupby(df.index.name)
        res = gb.apply(_as_what(what)) ## since first and last return NON NAN VALUES, we need to override them
    else:
        res = df
    res = res.drop(columns = _asof)
    if _columns in res.columns: ## we have a data frame with potentially mixed columns
        combos = list(set(res[_columns].values))
        if len(combos) == 1:
            columns = list(combos[0])
        else:
            columns = list(set(sum(combos, ())))
        res = res[columns]
    res.index.name = index_name
    if res.shape[1] == 1 and res.columns[0] == _series:
        res = res[_series]
    else:
        columns = [0 if c == _series else c for c in columns] ## this will clash with a dataframe with 0 column but thats fine
        res.columns = columns
    return res

def bi_asof(df, what = -1):
    """
    returns a timeseries stamped on the dates for which we have observations
    >>> df = pd.DataFrame(dict(a = [1,2,3,4,5,6], 
            _asof = drange(-2) + drange(-1) + drange(0)), 
            index = [dt(-2)] * 3 + [dt(-1)] * 2 + [dt(0)])
    >>> what = -1    

    """
    if not is_bi(df) or what == 'all':
        return df
    gb = df.sort_index().groupby(_asof)
    res = gb.apply(_as_what(what)) ## since first and last return NON NAN VALUES, we need to override them
    res = res.drop(columns = _asof)
    return res


def _drop_repeats(d):
    """ 
    This is applied to a dataframe where 
    - all the observed index values are the same.
    - the _asof column is already sorted
    
    If there are partial bad values, we ffill, to ensure, the new observation are not nan
    Conversely, if the old value is nan, we accept the new value
    
    1) we first drop repeated values
    2) if we have a NEW value but the SAME _asof, we keep the later value. 
    
    """
    no_asof = d.drop(columns = _asof).ffill()
    old_values = no_asof.iloc[:-1].values
    new_values = no_asof.iloc[1:].values
    repeats = new_values == old_values
    res = d[~np.concatenate([[False],repeats.min(axis=1)])]
    res = res.drop_duplicates(subset = [_asof], keep = 'last')
    return res
 

@loop(list)
def _get_columns(df):
    return tuple(sort([col for col in df.columns if col!=_asof]))


@loop(list)
def _add_columns(df):
    if _columns in df:
        return df
    else:
        df = df.copy()
        columns = tuple([col for col in df.columns if col!=_asof])
        df[_columns] = [columns]*len(df)
    return df
    
@loop(list)
def _set_unique_column(df, col = _series):
    df = df.copy()
    columns = [col if c!=_asof else c for c in df.columns]
    df.columns = columns    
    return df

def bi_merge(old_data, new_data, asof = 'now', existing_data = None):
    """
    merges two or more bitemporal tables.
    
    Parameters:
    -----------
    old_data: old values, can be None, dataframe or bitemporal
    
    new_data: dataframe or list of dataframes
    
    asof:
        policy for handling non-bitemporal new data
        If new_data is NOT bitemporal converts using asof
    
    existing_data:
        policy for handling old_data
        'ignore/overwrite': ignores old data
        False/0/None: ignore non-bitemporals
        other values: use these to convert    
    
    Example: new values override old ones on same index and asof
    --------
    >>> from pyg import * 
    >>> s = pd.Series([1,2,3], drange(-3,-1))
    >>> old_data = Bi(s, 0)    
    >>> new_data = Bi(s+1, 0)
    >>> asof = 'now'; existing_data = None
    >>> merged = bi_merge(old_data, new_data)
    >>> assert eq(merged, new_data)
    >>> assert eq(bi_read(merged), s+1)
    
    Example: handling of nan
    ------------------------
    >>> s = pd.Series([1,2,3], drange(-3,-1))
    >>> s1 = pd.Series([1,2,np.nan], drange(-3,-1))
    >>> s2 = pd.Series([1,np.nan,3], drange(-3,-1))
    >>> b1 = Bi(s1, 0)     
    >>> b2 = Bi(s2, 1)
    >>> merged = bi_merge(b1, b2)
    >>> assert eq(bi_read(merged), s)
    >>> assert eq(bi_read(merged, dt(-1)), s1)
    >>> assert eq(bi_read(merged, what = 0), s1)
    >>> assert eq(bi_read(merged, dt(0)), s)
    
    Example: handling of multiple column names with pseudo-series
    ------------------------------------------
    ## if we have SINGLE COLUMNS dataframes, even if they clash by name, we merge into a SERIES    
    >>> from pyg import * 
    >>> s = pd.Series([1,2,3], drange(-3,-1))
    >>> p1 = pd.DataFrame(dict(a = [1,2,4]), drange(-3,-1))
    >>> p2 = pd.DataFrame(dict(b = [1,3,4]), drange(-3,-1))
    >>> bs = Bi(s, 0)
    >>> bp1 = Bi(p1, 1)
    >>> bp2 = Bi(p2, 2)
    >>> m = bi_merge(bs, bp1)
    >>> assert sort(m.columns) == sort([_series, _asof])
    >>> m = bi_merge(bs, [bp1, bp2])    
    >>> assert sort(m.columns) == sort([_series, _asof])
    >>> assert eq(bi_read(m), as_series(p2))
    
    Example: handling of multiple column names with a 2d dataframe
    ------------------------------------------
    >>> from pyg import * 
    >>> from pyg_base._bitemporal import _asof, _series, _columns
    >>> s = pd.Series([1,2,3], drange(-2))
    >>> df = pd.DataFrame(dict(a = [1,2,4], b = [4,5,6]), drange(-2))
    >>> bs = Bi(s, dt(1))
    >>> bdf = Bi(df, dt(2))
    >>> m = bi_merge(bs, bdf)
    >>> assert sort(m.columns) == sort(['a','b',_asof, _series, _columns])
    >>> assert eq(bi_read(m, asof = dt(1)), s)
    >>> assert eq(bi_read(m, asof = dt(2)), df)
    >>> df2 = pd.DataFrame(dict(a = [1,2,4], b = [4,5,6]), drange(-1,1))
    >>> bdf2 = Bi(df2, dt(2))
    >>> m2 = bi_merge(bs, bdf2)
    >>> assert sort(bi_read(m2).columns) == sort(('b', 0, 'a'))

    """
    if existing_data in ('ignore', 'overwrite'):
        old_bis = []
    else:
        if not existing_data:
            old_bis = [b for b in as_list(old_data) if is_bi(b)]
        else:
            old_bis = [b if is_bi(b) else Bi(b, existing_data) for b in as_list(old_data)]
    new_bis = [b if is_bi(b) else Bi(b, asof) for b in as_list(new_data)]            
    bis = old_bis + new_bis
    if len(bis) == 0:
        return None
    if len(bis) == 1:
        return bis[0]
    bis = [b if is_bi(b) else Bi(b, asof) for b in bis]
    ### we first determine columns needed
    with_columns = [b for b in bis if _columns in b.columns]
    if len(with_columns):
        bis = _add_columns(bis)
    else:
        columns = list(set(_get_columns(bis)))
        if len(columns) > 1:
            ns = [len(c) for c in columns]
            if max(ns) > 1: ## too bad
                bis = _add_columns(bis)
            else: ## we remove all column names
                bis = _set_unique_column(bis)
    df = pd.concat(bis)
    index_name = df.index.name
    if index_name is None:
        df.index.name = 'index'
    gb = df.sort_values(_asof).groupby(df.index.name)
    res = pd.concat([_drop_repeats(d) for _, d in gb])
    res.index.name = index_name
    return res
    

def Bi(df, asof = None):
    """
    Creates a bitemporal dataframe
    
    :Example: make the _asof same as date index
    ---------
    >>> df = pd.DataFrame(dict(a = range(11)), drange(-10))
    >>> assert eq(Bi(df, 0)[_asof].values, df.index.values) 

    :Example: make the _asof right now...
    ---------
    >>> assert len(set(Bi(df)[_asof].values)) == 1

    :Example: make the _asof the 3am of next business...
    --------- 
    >>> Bi(df, ['1b', '3h'])
                 a               _asof
    2022-10-15   0 2022-10-18 03:00:00
    2022-10-16   1 2022-10-18 03:00:00
    ....
    2022-10-24   9 2022-10-25 03:00:00
    2022-10-25  10 2022-10-26 03:00:00

    :Example: make the _asof any exact date...
    ---------
    >>> asof = dt(2)
    >>> assert set(Bi(df, asof)[_asof]) ==  set([asof])
    
    """
    if asof is None or is_bi(df):
        return df
    if is_series(df):
        df = pd.DataFrame(df, columns = [_series])
    else:
        df = df.copy()
    if asof == 'shift':
        now = dt()
        df[_asof] = list(df.index[1:]) + [now]        
    elif is_bump(asof):
        now = dt()
        df[_asof] = dt_bump(df, asof).index
        df.loc[(df[_asof] > now), _asof] = now
    elif isinstance(asof, list):
        now = dt()
        df[_asof] = dt_bump(df, *asof).index
        df.loc[(df[_asof] > now), _asof] = now
    else:
        df[_asof] = dt(asof)
    return df




from pathlib import Path
import pandas as pd
from openpyxl import load_workbook

def pd_to_excel(filename, df, 
                sheet_name = None,
                if_sheet_exists = 'replace',
                date_format = None,
                datetime_format = None, 
                na_rep = '', 
                float_format = None, 
                columns = None,
                header = True,
                index_name = None,
                mode = None,
                suffix = '.xlsx'):
    """
    Alas, df.to_excel tends to overwrite existing files completely.
    - pd_to_excel defaults to overwriting just the sheet we write
    - supports appending to sheet rather than replace
    """
    
    if index_name:
        df = df.copy()
        df.index.name = index_name
    
    fname = Path(filename).with_suffix(suffix)
    file_exists = fname.is_file()
    if mode is None:
        mode = 'a' if file_exists else 'w'
    
    wb = None
    if sheet_name is None:
        if file_exists:
            wb = load_workbook(fname)
            sheet_names = wb.sheetnames
            sheet_name = sheet_names[0]
        else:
            sheet_name = 'Sheet1'
    
    ## fake append
    if if_sheet_exists == 'append' : #fake append
        if file_exists:
            if wb is None:
                wb = load_workbook(fname)
                sheet_names = wb.sheetnames
            sheet_exists = sheet_name in sheet_names
        if sheet_exists:
            old = pd.read_excel(fname, sheet_name = sheet_name)
            df = pd.concat([old, df])
        if_sheet_exists = 'replace'
    
    with pd.ExcelWriter(fname, engine = 'openpyxl',
                        mode = mode,
                        date_format = date_format,
                        datetime_format = datetime_format,
                        if_sheet_exists = if_sheet_exists) as writer:
        df.to_excel(writer, sheet_name = sheet_name, na_rep = na_rep,
                    float_format = float_format, columns = columns, header = header)
    
    return df
        


from _collections_abc import dict_keys, dict_values
from pyg_base._as_list import as_list, as_tuple
from pyg_base._as_float import as_float
from pyg_base._dict import Dict
from pyg_base._zip import zipper, lens
from pyg_base._types import is_str, is_strs, is_arr, is_df, is_dicts, is_int, is_ints, is_tuple, is_bools, is_nan
from pyg_base._tree import is_tree, tree_to_table
from pyg_base._inspect import getargs
from pyg_base._decorators import kwargs_support, try_none, try_back
from pyg_base._dates import ndt
from pyg_base._sort import sort, cmp
from pyg_base._file import read_csv
from pyg_base._cache import cache
from pyg_base._logger import logger
from pyg_npy import pd_read_npy
from functools import reduce

import pandas as pd
import re

Pattern = type(re.compile('')) ## this types definition moves across re. versions

__all__ = ['dict_concat', 'dictable', 'is_dictable']

@cache
def _print_cols(*lcols):
    logger.info('inner joining on %s. To stop message, specify columns explicitly in the join'%list(lcols))

    
def nan2none(v):
    return None if is_nan(v) or (is_str(v) and (len(v.strip()) == 0 or v[0] == '#' or v == 'n/a')) else v

def relabel_lower(v):
    if is_str(v):
        return v.strip().replace(' ', '_').replace('"','').replace("'",'').replace(',','').lower()
    else:
        return v    
    
def dict_concat(*dicts):
    """
    A method of turning a list of dicts into one happy dict sharing its keys

    :Parameters:
    ----------------
    *dicts : list of dicts
        a list of dicts

    :Returns:
    -------
    dict
        a dict of a list of values.
    
    :Example:
    --------------
    >>> dicts = [dict(a=1,b=2), dict(a=3,b=4,c=5)]
    >>> assert dict_concat(dicts) == dict(a = [1,3], b = [2,4], c = [None,5])
    >>> dicts = [dict(a=1,b=2)]
    >>> assert dict_concat(dicts) == dict(a = [1], b = [2])
    >>> assert dict_concat([]) == dict()
    """
    
    dicts = as_list(dicts)
    if len(dicts) == 0:
        return {}
    elif len(dicts) == 1:
        return {key: [value] for key, value in dicts[0].items()}  # a shortcut, we dont need to run a full merge
    possible_keys = list(set([tuple(sorted(d.keys())) for d in dicts]))
    if len(possible_keys) == 1:
        pairs = [sorted(d.items()) for d in dicts]
        keys = possible_keys[0]
        values = zip(*[[value for _, value in row] for row in pairs])
        res = dict(zip(keys, map(list, values)))
        return res
    else:
        keys = reduce(lambda res, keys: res | set(keys), possible_keys, set())
        return {key: [d.get(key) for d in dicts] for key in keys}
    

def _text_box(value, max_rows = 5, max_chars = 50):
    """
    a = dictable(a = range(10)) * dict(b = range(10))
    b = (dictable(a = range(5)) * dict(b = range(5))) * dict(c = range(10))
    self = c = dictable(key = ['a', 'b'] * 5, rs = [a,b]*5)
    """
    v = str(value)
    res = v.split('\n')
    if max_rows:
        res = res[:max_rows]
    if max_chars:
        res = [row[:max_chars] for row in res]
    return res


def _data_columns_as_dict(data, columns = None):
    """
    >>> assert _data_columns_as_dict(data = [], columns = []) == dict()
    >>> assert _data_columns_as_dict([], 'a') == dict(a = [])
    >>> assert _data_columns_as_dict(data = [], columns = ['a','b']) == dict(a = [], b = [])
    
    :Parameters:
    ----------------
    data : TYPE
        DESCRIPTION.
    columns : TYPE, optional
        DESCRIPTION. The default is None.

    :Returns:
    -------
    TYPE
        DESCRIPTION.

    """
    if isinstance(data, zip):
        data = list(data)
    if data is None or (isinstance(data, list) and data == []):
        return {}
    elif isinstance(data, dict):
        return dict_concat(tree_to_table(data, columns)) if is_tree(columns) else dict(data)
    if columns is not None:
        if is_str(columns):
            if (is_str(data) and '.xls' in data) or isinstance(data, pd.io.excel.ExcelFile):
                return dict(dictable.read_excel(data, columns))
            else:
                return {columns : data}
        else:
            return dict(zipper(columns, zipper(*data)))
    else:
        if is_str(data):
            if data.endswith('.csv'):
                data = read_csv(data)
            elif '.xls' in data:
                data = dict(dictable.read_excel(data))
            elif data.endswith('.parquet'):
                data = pd.read_parquet(data)
            elif data.endswith('.pickle'):
                data = pd.read_pickle(data)
            elif data.endswith('.dictable'):
                try:
                    data = pd.read_pickle(data)
                except Exception:
                    data = pd.read_parquet(data)
            elif data.endswith('.npy'):
                data = pd_read_npy(data)
            else:
                return dict(data = data)
        if is_df(data):
            if data.index.name is not None:
                data = data.reset_index()
            return data.to_dict('list')
        tp = str(type(data)).lower()        
        if hasattr(data, 'next') and 'cursor' in tp:
            data = list(data)
        elif hasattr(data, 'find') and 'collection' in tp:
            data = list(data.find({}))
        if isinstance(data, list):
            if len(data) == 0:
                return {}
            elif min([is_tuple(row) and len(row) == 2 for row in data]):
                return dict(data)
            elif is_dicts(data):
                return dict_concat(data)
            elif min([isinstance(i, list) for i in data]):
                return dict(zipper(data[0], zipper(*data[1:])))
            else:
                return dict(data = data)
        elif isinstance(data, dict):
            return dict(data)
        else:
            return dict(data = data)


def _value(value):
    if value is None:
        return [None]
    elif isinstance(value, (dict_values, dict_keys, range)):
        return list(value)
    else:
        return as_list(value)

class dictable(Dict):
    """
    :What is dictable?:
    -------------------
    dictable is a table, a collection of iterable records. It is also a dict with each key being a column. 
    Why not use a pandas.DataFrame? pd.DataFrame leads a dual life: 
        - by day an index-based optimized numpy array supporting e.g. timeseries analytics etc.
        - by night, a table with keys supporting filtering, aggregating, pivoting on keys as well as inner/outer joining on keys.

    
    dictable only tries to do the latter. dictable should be thought of as a 'container for complicated objects' rather than just an array of primitive floats.
    In general, each cell may contain timeseries, yield_curves, machine-learning experiments etc.
    The interface is very succinct and allows the user to concentrate on logic of the calculations rather than boilerplate.
    
    dictable supports quite a flexible construction:

    :Example: construction using records
    ------------------------------------
    >>> from pyg import *; import pandas as pd
    >>> d = dictable([dict(name = 'alan', surname = 'atkins', age = 39, country = 'UK'), 
    >>>               dict(name = 'barbara', surname = 'brown', age = 29, country = 'UK')])
    
    :Example: construction using columns and constants
    ---------------------------------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    
    :Example: construction using pandas.DataFrame
    ---------------------------------------------
    >>> original = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> df_from_dictable = pd.DataFrame(original)
    >>> dictable_from_df = dictable(df_from_dictable)
    >>> assert original == dictable_from_df

    :Example: construction rows and columns
    ---------------------------------------
    >>> d = dictable([['alan', 'atkins', 39, 'UK'], ['barbara', 'brown', 29, 'UK']], columns = ['name', 'surname', 'age', 'country'])


    :Access: column access
    ----------------------
    >>> assert d.keys() ==  ['name', 'surname', 'age', 'country']
    >>> assert d.name == ['alan', 'barbara']
    >>> assert d['name'] == ['alan', 'barbara']
    >>> assert d['name', 'surname'] == [('alan', 'atkins'), ('barbara', 'brown')]
    >>> assert d[lambda name, surname: '%s %s'%(name, surname)] == ['alan atkins', 'barbara brown']


    :Access: row access & iteration
    -------------------------------
    >>> assert d[0] == {'name': 'alan', 'surname': 'atkins', 'age': 39, 'country': 'UK'}
    >>> assert [row for row in d] == [{'name': 'alan', 'surname': 'atkins', 'age': 39, 'country': 'UK'},
    >>>                               {'name': 'barbara', 'surname': 'brown', 'age': 29, 'country': 'UK'}]

    Note that members access is commutative: 

    >>> assert d.name[0] == d[0].name == 'alan'
    >>> d[lambda name, surname: name + surname][0] == d[0][lambda name, surname: name + surname]
    >>> assert sum([row for row in d], dictable()) == d

    :Example: adding records
    ------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> d = d + {'name': 'charlie', 'surname': 'chocolate', 'age': 49} # can add a record directly
    >>> assert d[-1] == {'name': 'charlie', 'surname': 'chocolate', 'age': 49, 'country': None}
    >>> d += dictable(name = ['dana', 'ender'], surname = ['deutch', 'esterhase'], age = [10, 20], country = ['Germany', 'Hungary'])
    >>> assert d.name == ['alan', 'barbara', 'charlie', 'dana', 'ender']
    >>> assert len(dictable.concat([d,d])) == len(d) * 2

    :Example: adding columns
    ------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    
    >>> ### all of the below are ways of adding columns ####
    >>> d.gender == ['m', 'f']
    >>> d = d(gender = ['m', 'f'])
    >>> d['gender'] == ['m', 'f']
    >>> d2 = dictable(gender = ['m', 'f'], profession = ['astronaut', 'barber'])
    >>> d = d(**d2)

    :Example: adding derived columns
    --------------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> d = d(full_name = lambda name, surname: proper('%s %s'%(name, surname))) 
    >>> d['full_name'] = d[lambda name, surname: proper('%s %s'%(name, surname))]
    >>> assert d.full_name == ['Alan Atkins', 'Barbara Brown']

    :Example: dropping columns
    ---------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> del d.country # in place
    >>> del d['age'] # in place
    >>> assert (d - 'name')[0] ==  {'surname': 'atkins'} and d[0] == {'name': 'alan', 'surname': 'atkins'}

    :Example: row selection, inc/exc
    --------------------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> assert len(d.exc(name = 'alan')) == 1
    >>> assert len(d.exc(lambda age: age<30)) == 1 # can filter on *functions* of members, not just members.
    >>> assert d.inc(name = 'alan').surname == ['atkins']
    >>> assert d.inc(lambda age: age<30).name == ['barbara']
    >>> assert d.exc(lambda age: age<30).name == ['alan']

    dictable supports:
        - sort 
        - group-by/ungroup
        - list-by/ unlist
        - pivot/unpivot
        - inner join, outer join and xor

    Full details are below.
    """
    def __init__(self, data = None, columns = None, **kwargs):
        kwargs = {key :_value(value) for key, value in kwargs.items()}
        data_kwargs = {key: _value(value) for key, value in _data_columns_as_dict(data, columns).items()}
        kwargs.update(data_kwargs)
        if is_strs(columns) and (len(data_kwargs) == 0 or not is_str(columns)):
            kwargs = {key : kwargs.get(key, [None]) for key in columns} if len(kwargs)>0 else {key : [] for key in columns}
        n = lens(*kwargs.values())
        kwargs = {str(key) if is_int(key) else key : value * n if len(value)==1 else value for key, value in kwargs.items()}
        super(dictable, self).__init__(kwargs)
        
    _dict = Dict

    def __len__(self):
        return lens(*self.values())
    
    @property
    def shape(self):
        return (len(self), len(self.keys()))
    
    @property
    def columns(self):
        return self.keys()

    def get(self, key, default = None):
        if key in self:
            return self[key]
        else:
            return [default] * len(self)
        
    def __iter__(self):
        for row in zip(*self.values()):
            yield self._dict(zip(self.keys(), row))
            
    def update(self, other):
        for k, v in other.items():
            self[k] = v
    
    def __setitem__(self, key, value):
        n = len(self)
        value = _value(value)
        if len(value) == n or len(self.keys()) == 0:
            pass
        elif len(value) == 1:
            value = value * n 
        else:
            raise ValueError('cannot set item of length %s in table of length %s'%(len(value), n))
        super(dictable, self).__setitem__(str(key) if is_int(key) else key, value)
    
    def __getitem__(self, item):
        if is_arr(item) and len(item.shape) == 1:
            item = list(item)
        if isinstance(item, slice):
            return type(self)({key : value[item] for key, value in self.items()})
        if isinstance(item, (dict_keys, dict_values, range)):
            item = list(item)
        if isinstance(item , list):
            if len(item) == 0:
                return type(self)(data = [], columns = self.keys())
            elif is_strs(item):
                return type(self)(super(dictable, self).__getitem__(item))
            elif is_bools(item):
                res = type(self)([row for row, tf in zipper(list(self), item) if tf])
                return res if len(res) else type(self)([], self.keys())
            elif is_ints(item):
                values = list(zip(*self.values()))
                return type(self)(data = [values[i] for i in item], columns = self.keys())
            else:
                raise ValueError('We dont know how to understand this item %s'%item)
        elif is_int(item):
            return self._dict({key : value[item] for key, value in self.items()})
        elif item in self.keys():
            return super(dictable, self).__getitem__(item)
        elif is_tuple(item):
            return list(zip(*[self[i] for i in item]))
        elif callable(item):
            return self.apply(item)
        else:
            raise KeyError('item %s not found'%item)

    def __getattr__(self, attr):
        if attr in self:
            return self[attr]
        elif attr.startswith('_'):
            return super(dictable, self).__getattr__(attr)
        elif attr.startswith('find_'):
            key = attr[5:]
            if key not in self:
                raise KeyError('cannot find %s in dictable'%key)
            def f(*args, **kwargs):
                items = self.inc(*args, **kwargs)
                if len(items) == 0:
                    raise ValueError('no %s found'%key)
                item = items[key]
                if len(item)>1:
                    item = list(set(item))
                    if len(item)>1:
                        raise ValueError('multiple %s found %s'%(key, item))
                return item[0]
            f.__name__ = 'find_%s'%key
            return f
        else:
            raise AttributeError('%s not found'%attr)
 
    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            super(dictable, self).__setattr__(attr, value)
        else:
            self.__setitem__(attr, value)

    def __delattr__(self, attr):
        if attr.startswith('_'):
            super(dictable, self).__delattr__(attr)
        else:
            super(dictable, self).__delitem__(attr)
    
    # def _encode(self):
    #     """
    #     We override the encoding method for dictable by enforcing columns to be specified as part of the dict.
    #     This ensures that when constructed from mongodb, old columns that may still be part of the cell are not included in the construction

    #     """
    #     res = {k : _encode(v) for k, v in self.items()}
    #     if _obj not in res:
    #         res[_obj] = _encode(type(self))
    #     res['columns'] = self.columns
    #     return res

    def inc(self, *functions, **filters):
        """
        performs a filter on what rows to include

        :Parameters:
        ----------------
        *functions : callables or a dict
            filters based on functions of each row
        **filters : value or list of values
            filters per each column

        :Returns:
        -------
        dictable
            table with rows that satisfy all conditions.
            
        
        :Example: filtering on keys
        -------
        >>> from pyg import *; import numpy as np
        >>> d = dictable(x = [1,2,3,np.nan], y = [0,4,3,5])
        >>> assert d.inc(x = np.nan) == dictable(x = np.nan, y = 5)            
        >>> assert d.inc(x = 1) == dictable(x = 1, y = 0)            
        >>> assert d.inc(x = [1,2]) == dictable(x = [1,2], y = [0,4]) 

        :Example: filtering on regex
        -------
        >>> import re
        >>> d = dictable(text = ['once', 'upon', 'a', 'time', 'in', 'the', 'west', 1, 2, 3])
        >>> assert d.inc(text = re.compile('o')) == dictable(text = ['once', 'upon'])
        >>> assert d.exc(text = re.compile('e')) == dictable(text = ['upon', 'a', 'in', 1, 2, 3])
        
        
        :Example: filtering on callables
        --------------
        >>> from pyg import *; import numpy as np
        >>> d = dictable(x = [1,2,3,np.nan], y = [0,4,3,5])
        >>> assert d.inc(lambda x,y: x>y) == dictable(x = 1, y = 0)

        """
        res = self.copy()
        if len(functions) + len(filters) == 0:
            return res
        functions = as_list(functions)
        for function in functions:
            if type(function) == dict:
                filters.update(function)
            else:
                f = kwargs_support(function)
                res = type(self)([row for row in res if f(**row)])
        for key, value in filters.items():
            if value is None:
                res = res[[r is None for r in res[key]]]
            elif is_nan(value):
                res = res[[is_nan(r) for r in res[key]]]
            elif isinstance(value, Pattern):
                res = res[[is_str(r) and value.search(r) is not None for r in res[key]]]                
            else:
                value = as_list(value)
                res = res[[r in value for r in res[key]]]
        if len(res) == 0:
            return type(self)([], self.keys())
        return res                

    def exc(self, *functions, **filters):
        """
        performs a filter on what rows to exclude

        :Parameters:
        ----------------
        *functions : callables or a dict
            filters based on functions of each row
        **filters : value or list of values
            filters per each column

        :Returns:
        -------
        dictable
            table with rows that satisfy all conditions excluded.
            
        
        :Example: filtering on keys
        -------
        >>> from pyg import *; import numpy as np
        >>> d = dictable(x = [1,2,3,np.nan], y = [0,4,3,5])
        >>> assert d.exc(x = np.nan) == dictable(x = [1,2,3], y = [0,4,3])         
        >>> assert d.exc(x = 1) == dictable(x = [2,3,np.nan], y = [4,3,5])
        >>> assert d.exc(x = [1,2]) == dictable(x = [1,2], y = [0,4]) 

        :Example: filtering on callables
        --------------
        >>> from pyg import *; import numpy as np
        >>> d = dictable(x = [1,2,3,np.nan], y = [0,4,3,5])
        >>> assert d.exc(lambda x,y: x>y) == dictable(x = 1, y = 0)
        
        """
        res = self.copy()
        if len(functions) + len(filters) == 0:
            return res
        functions = as_list(functions)
        for function in functions:
            if type(function) == dict:
                filters.update(function)
            else:
                f = kwargs_support(function)
                res = type(self)([row for row in res if not f(**row)])
        for key, value in filters.items():
            pass
            if value is None:
                res = res[[r is not None for r in res[key]]]
            elif is_nan(value):
                res = res[[not is_nan(r) for r in res[key]]]
            elif isinstance(value, Pattern):
                res = res[[(not is_str(r)) or value.search(r) is None for r in res[key]]]                
            else:
                value = as_list(value)
                res = res[[r not in value for r in res[key]]]
        if len(res) == 0:
            return type(self)([], self.keys())
        return res                


    def if_none(self, none = None, **kwargs):
        """
        runs a column calculation if a column is not there, or if the existing value is considered missing.
        
        Example:
        --------
        >>> rs = dictable(a = [1,'hello'])
        >>> rs = rs(b = try_none(lambda a: a[0]))  ## fails for numerical value
        >>> rs

        >>> dictable[2 x 2]
        >>> a    |b   
        >>> 1    |None
        >>> hello|h   
        
        what shall we do with integer values for which b is None?
        
        >>> rs = rs.if_none(b = lambda a: a+3)
        >>> rs
        >>> dictable[2 x 2]
        >>> a    |b   
        >>> 1    |4
        >>> hello|h   


        Example:
        -------- 
        >>> rs = dictable(a = range(3))
        >>> rs = rs(b = try_nan(lambda a: 10/a))
        >>> rs = rs.if_none(np.nan, b = 'sorry no can do')

        >>> dictable[3 x 2]
        >>> a|b              
        >>> 0|sorry no can do
        >>> 1|10.0           
        >>> 2|5.0        
        
        """
        if none is None:
            check = lambda value: value is None
        elif not callable(none):
            check = lambda value: value in as_list(none)
        else:
            check = none
        res = self
        for key, value in kwargs.items():
            if key not in res.keys():
                res[key] = res[value]
            else:
                if callable(value):
                    res[key] = [row[key] if not check(row[key]) else row[value] for row in res]
                else:
                    res[key] = [row[key] if not check(row[key]) else value for row in res]                    
        return res

    
    def apply(self, function, **default_params):
        f = kwargs_support(function)
        return [f(**(default_params|row)) for row in self]

    def do(self, function, *keys):
        """
        applies a function(s) on multiple keys at the same time

        :Parameters:
        ----------------
        function : callable or list of callables
            function to be applied per column
        *keys : string/list of strings
            list of columns to be applied. If missing, applied to all columns

        :Returns:
        -------
        res : dictable

        :Example:
        --------------
        >>> from pyg import *
        >>> d = dictable(name = ['adam', 'barbara', 'chris'], surname = ['atkins', 'brown', 'cohen'])
        >>> assert d.do(proper) == dictable(name = ['Adam', 'Barbara', 'Chris'], surname = ['Atkins', 'Brown', 'Cohen'])

        :Example: using another column in the calculation
        -------
        >>> from pyg import *
        >>> d = dictable(a = [1,2,3,4], b = [5,6,9,8], denominator = [10,20,30,40])
        >>> d = d.do(lambda value, denominator: value/denominator, 'a', 'b')
        >>> assert d == dictable(a = 0.1, b = [0.5,0.3,0.3,0.2], denominator = [10,20,30,40])
        
        """
        res = self.copy()
        keys = as_list(keys)
        if len(keys)  == 0:
            keys = self.keys()
        for key in keys:    
            for f in as_list(function):
                args = as_list(try_none(getargs)(f))
                res[key] = [f(row[key], **{k : v for k, v in row.items() if k in args[1:]}) for row in res]
        return res

    @classmethod
    def concat(cls, *others):
        """
        adds together multiple dictables. equivalent to sum(others, self) but a little faster

        :Parameters:
        ----------------
        *others : dictables
            records to be added to current table

        :Returns:
        -------
        merged : dictable
            sum of all records
            
        :Example:
        -------
        >>> from pyg import *
        >>> d1 = dictable(a = [1,2,3])
        >>> d2 = dictable(a = [4,5,6])
        >>> d3 = dictable(a = [7,8,9])
        
        >>> assert dictable.concat(d1,d2,d3) == dictable(a = range(1,10))
        >>> assert dictable.concat([d1,d2,d3]) == dictable(a = range(1,10))
        """
        others = as_list(others)
        others = [cls(other) if not isinstance(other, cls) else other for other in others]
        if len(others) == 0:
            return cls()
        elif len(others) == 1:
            return others[0]
        concated = dict_concat(others)
        merged = cls({key : sum(value, []) for key, value in concated.items()}) 
        return merged

    def __repr__(self):
        return 'dictable[%s x %s]'%self.shape + '\n%s'%self.__str__(3, 150)
    
    
    def __str__(self, max_rows = None, max_width = 150):
        return self.to_string(cat = max_rows, mid = '...%i rows...'%len(self), max_width=max_width)
    
    def to_string(self, colsep = '|', rowsep = '', cat = None, max_rows = 5, max_width = None, mid = '...', header = None, footer = None):
        """
        convrts dictable to string
        :Example:
        ---------
        >>> colsep = '|'; rowsep = ''; cat = None; max_rows = 5; max_width = None; mid = '...'; header = None; footer = None
        
        """
        if cat and len(self) > 2 * cat:
            box = {key: [_text_box(key)] + 
                        [_text_box('dictable[%s x %s]\n%s'%(len(v), len(v.keys()), v.keys()) if isinstance(v, dictable) else v, max_rows) for v in value[:cat]+value[-cat:]]
                    for key, value in self.items()}
        else:
            box = {key: [_text_box(key)] + 
                        [_text_box('dictable[%s x %s]\n%s'%(len(v), len(v.keys()), v.keys()) if isinstance(v, dictable) else v, max_rows) for v in value]
                    for key, value in self.items()}
    
        chars = {key : max([max([len(r) for r in v]) for v in value]) for key, value in box.items()}
        padded = {key : [[r + ' ' * (chars[key]-len(r)) for r in v] for v in value] for key, value in box.items()}
        def text(padded, keys = None):
            padded_ = padded if keys is None else {key: padded[key] for key in keys}
            rows = list(zip(*padded_.values()))
            empty_rows = [[' '*chars[key]] for key in padded_]
            ns = [max([len(v) for v in row]) for row in rows]
            if len(rowsep) == 1:
                sep_rows = [[rowsep*chars[key]] for key in padded_]
                res = ['\n'.join([colsep.join(v) for v in zip(*[r + e * (n-len(r)) + s for r,e,s in zip(row, empty_rows, sep_rows)])]) 
                       for row,n in zip(rows, ns)]
            else:
                res = ['\n'.join([colsep.join(v) for v in zip(*[r+e*(n-len(r)) for r, e in zip(row, empty_rows)])]) for row, n in zip(rows, ns)]
            if mid and cat and len(self)>2*cat:
                n = (len(res)+1)//2
                res = res[:n] + [mid] + res[n:]
            if rowsep == 'header':
                res = res[:1] + ['+'.join(['-'*chars[key] for key in padded_])] + res[1:]
            if header:
                res = [header] + res
            if footer:
                res = res + [footer]
            return '\n'.join(res)
        if max_width is None or max_width<=0:
            return text(padded)
        else:
            keys = list(padded.keys())
            res = []
            while len(keys)>0:
                i = 0
                width = 0
                while width<max_width and i<len(keys):
                    width+= chars[keys[i]]
                    i+=1
                res.append(text(padded, keys[:i]))
                keys = keys[i:]
            return '\n\n'.join(res)

    def sort(self, *by, **byval):
        """
        Sorts the table either using a key, list of keys or functions of members. Also allows sort on specific orders of the keys
        
        
        :Example:
        -------
        >>> import numpy as np
        >>> self = dictable(a = [_ for _ in 'abracadabra'], b=range(11), c = range(0,33,3))
        >>> self.d = list(np.array(self.c) % 11)
        >>> res = self.sort('a', 'd')
        >>> assert list(res.c) == list(range(11))
        
                
        >>> d = dictable(a = ['a', 1, 'c', 0, 'b', 2]).sort('a')        
        >>> res = d.sort('a','c')
        >>> print(res)
        >>> assert ''.join(res.a) == 'aaaaabbcdrr' and list(res.c) == [0,4,8,9,10] + [2,3] + [1] + [7] + [5,6]
        
        >>> d = d.sort(lambda b: b*3 % 11) ## sorting again by c but using a function
        >>> assert list(d.c) == list(range(11))
        
        :Example: Sorting based on specific values
        ---------
        >>> rs = dictable(key = list('abcdeffg'), gender = list('mmffmmff'))
        >>> rs = rs.sort(key = ['c', 'a', 'f', 'd'], gender = ['f', 'm'])
        
        key|gender
        ----------
        c  |f     
        a  |m     
        f  |f     <--- female f before male f
        f  |m     <---
        d  |f     
        g  |f     <--- g,b,e entries with no first key provided in sort, so move to bottom of list
        b  |m     
        e  |m    
        """
        if len(self) == 0:
            return self.copy()
        elif len(by):
            keys = self[by]            
        elif len(byval):
            dicts = {k : dict(zip(vals, range(len(vals)))) for k, vals in byval.items()}
            keys = [[d.get(row[k], len(d)) for k,d in dicts.items()] for row in self]            
        else:
            return self.copy()
        keys2id = list(zip(keys, range(len(self))))
        _, rows = zip(*sort(keys2id))
        return type(self)({key: [value[i] for i in rows] for key, value in self.items()})

    def __add__(self, other):
        if other in (None, 0):
            return self
        return self.concat(self, other)

    def __and__(self, other):
        """
        equivalent 
        
        Example:
        --------
        >>> from pyg import * 
        >>> rs = dictable(a = 1, b = 2, c = [3,4])
        >>> assert rs & ['c', 'd'] == rs[['c']]

        """
        return self[self.columns & other]
            
    __radd__ = __add__
    
    def _listby(self, by):
        keys = self[by]
        keys2id = sort(list(zip(keys, range(len(self)))))
        prev = None
        res = []
        row = []
        for key, i in keys2id:
            if len(row) == 0 or key==prev:
                row.append(i)
            else:
                res.append((prev, row))
                row = [i]
            prev = key 
        res.append((prev, row))
        return zip(*res)

    def listby(self,*by):
        if len(self) == 0:
            return self.copy()
        if len(by) == 0:
            by = self.keys()
        by = as_tuple(by)
        if len(by) == 0:
            return type(self)({key: [value] for key, value in dict(self).items()})
        xs, ids = self._listby(by)
        rtn = type(self)(xs, by)
        rtn.update({k: [[self[k][i] for i in y] for y in ids] for k in self.keys() if k not in by})
        return rtn
    
    @classmethod
    def read_excel(cls, io, sheet_name = None, floats = None, ints = None, dates = None, no_none = None):
        if is_str(io):
            io = io.lower()
            if '!' in io and sheet_name is None: ## handling input like: ExcelFile.xlsx!SheetName
                io, sheet_name = io.split('!')
            if not '.xls' in io:
                io = io + '.xlsx'
            io = pd.ExcelFile(io)
        if sheet_name is None:
            sheet_name = 0
        df = io.parse(sheet_name)
        res = cls(df)
        res = res.relabel(lambda v: v.strip().replace(' ', '_').replace('"','').replace("'",''))
        res = res.do(nan2none)
        if floats:
            res = res.do(try_back(as_float), floats)
        if ints:
            res = res.do(try_back(int), ints)
        if dates:
            res = res.do(try_back(ndt), dates)
        if no_none:
            for col in as_list(no_none):
                res = res.exc(**{col : None})
        for col in [col for col in res.keys() if col.startswith('Unnamed:')]:
            if set(res[col]) == set([None]):
                res = res-col
        return res
    
    def unlist(self):
        """
        undoes listby...
        
        :Example:
        -------
        >>> x = dictable(a = [1,2,3,4], b= [1,0,1,0])
        >>> x.listby('b')
        
        dictable[2 x 2]
        b|a     
        0|[2, 4]
        1|[1, 3]
        
        
        >>> assert x.listby('b').unlist().sort('a') == x

        :Returns:
        -------
        dictable
            a dictable where all rows with list in them have been 'expanded'.

        """
        return self.concat([row for row in self]) if len(self) else self
    
    def groupby(self,*by, grp = 'grp'):
        """
        Similar to pandas groupby but returns a dictable of dictables with a new column 'grp' 
        
        :Example:
        -------
        >>> x = dictable(a = [1,2,3,4], b= [1,0,1,0])
        >>> res = x.groupby('b')
        >>> assert res.keys() == ['b', 'grp']
        >>> assert is_dictable(res[0].grp) and res[0].grp.keys() == ['a']

        :Parameters:
        ----------------
        *by : str or list of strings
        
            gr.
        grp : str, optional
            The name of the column for the dictables per each key. The default is 'grp'.

        :Returns:
        -------
        dictable
            A dictable containing the original keys and a dictable per unique key.

        """
        if len(self) == 0:
            return self.copy()
        if len(by) == 0:
            by = self.keys()
        by = as_tuple(by)
        if len(by) == 0:
            raise ValueError('cannot groupby on no keys, left with original dictable')
        elif len(by) == len(self.keys()):
            raise ValueError('cannot groupby on all keys... nothing left to group')
        xs,ys = self._listby(by)
        rtn = type(self)(xs, by)
        rtn[grp] = [type(self)({k: [self[k][i] for i in y] for k in self.keys() if k not in by}) for y in ys]
        return rtn

    def ungroup(self, grp = 'grp'):
        """
        Undoes groupby

        :Example:
        -------
        >>> x = dictable(a = [1,2,3,4], b= [1,0,1,0])
        >>> self = x.groupby('b')
        
        :Parameters:
        ----------------
        grp : str, optional
            column name where dictables are. The default is 'grp'.

        :Returns:
        -------
        dictable.

        """
        return self.concat([row.pop(grp)(**row.do(lambda v: [v])) for row in self])
        
    
    def join(self, other, lcols = None, rcols = None, mode = None):
        """
        
        Performs either an inner join or a cross join between two dictables

        :Example: inner join
        -------------------------------
        >>> from pyg import *
        >>> x = dictable(a = ['a','b','c','a']) 
        >>> y = dictable(a = ['a','y','z'])
        >>> assert x.join(y) == dictable(a = ['a', 'a'])

        :Example: outer join
        -------------------------------
        >>> from pyg import *
        >>> x = dictable(a = ['a','b']) 
        >>> y = dictable(b = ['x','y'])
        >>> assert x.join(y) == dictable(a = ['a', 'a', 'b', 'b'], b = ['x', 'y', 'x', 'y'])

        """
        _lcols = lcols
        if not isinstance(other, dictable):
            other = dictable(other)
        if lcols is None:
            lcols = self.keys() & other.keys()
        if rcols is None:
            rcols = lcols
        lcols = as_tuple(lcols); rcols = as_tuple(rcols)
        if len(lcols)!=len(rcols):
            raise ValueError('cannot inner join when cols on either side mismatch in length %s vs %s'%(lcols, rcols))
        elif _lcols is None:
            if len(lcols) > 0:
                _print_cols(*lcols)
        cols = []
        for lcol, rcol in zip(lcols, rcols):
            if is_str(lcol):
                cols.append(lcol)
            elif is_str(rcol):
                cols.append(rcol)
            else:
                raise ValueError('Cannot use a formula to inner join on both left and right %s %s'%(lcol, rcol))
        lkeys = self.keys() - cols
        rkeys = other.keys() - cols
        jkeys = lkeys & rkeys
        lkeys = lkeys - jkeys
        rkeys = rkeys - jkeys
        
        if len(cols):
            lxs, lids = self._listby(lcols)
            rxs, rids = other._listby(rcols)
            ls = len(lxs)
            rs = len(rxs)
            l = 0
            r = 0
            res = []
            while l<ls and r<rs:
                while l<ls and r<rs and cmp(lxs[l],rxs[r]) == -1:
                    l+=1
                while l<ls and r<rs and cmp(lxs[l],rxs[r]) == 1:
                    r+=1
                if l<ls and r<rs and lxs[l] == rxs[r]:
                    res.append((lxs[l], lids[l], rids[r]))
                    r+=1
                    l+=1
            if len(res) == 0:
                return type(self)([], cols + lkeys + rkeys + jkeys)
            xs, lids, rids = zip(*res)
            ns = [len(l)*len(r) for l, r in zip(lids, rids)]
            rtn = type(self)(sum([[x]*n for x,n in zip(xs,ns)], []), cols)
        else:
            rtn = type(self)()
            lids = [range(len(self))]; rids = [range(len(other))]
        for k in lkeys:
            v= self[k]
            rtn[k] = sum([[v[l] for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])
        for k in rkeys:
            v= other[k]
            rtn[k] = sum([[v[r] for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])            
        for k in jkeys:
            if (is_str(mode) and mode[0].lower() == 'l') or mode == 0:
                v = self[k]
                rtn[k] = sum([[v[l] for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])
            elif (is_str(mode) and mode[0].lower() == 'r') or mode == 1:
                v = other[k]
                rtn[k] = sum([[v[r] for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])
            elif callable(mode):
                lv = self[k]
                rv = other[k]
                rtn[k] = sum([[mode(lv[l], rv[r]) for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])            
            else:                
                lv = self[k]
                rv = other[k]
                rtn[k] = sum([[(lv[l], rv[r]) for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])            
        return rtn
    
    def xor(self, other, lcols = None, rcols = None, mode = 'l'):
        """
        returns what is in lhs but NOT in rhs (or vice versa if mode = 'r'). Together with inner joining, can be used as left/right join

        :Examples:
        --------------
        >>> from pyg import *
        >>> self = dictable(a = [1,2,3,4])
        >>> other = dictable(a = [1,2,3,5])
        >>> assert self.xor(other) == dictable(a = 4) # this is in lhs but not in rhs
        >>> assert self.xor(other, lcols = lambda a: a * 2, rcols = 'a') == dictable(a = [2,3,4]) # fit can be done using formulae rather than actual columns

        The XOR functionality can be performed using quotient (divide):
        >>> assert lhs/rhs == dictable(a = 4)
        >>> assert rhs/lhs == dictable(a = 5)

        >>> rhs = dictable(a = [1,2], b = [3,4])
        >>> left_join_can_be_done_simply_as = lhs * rhs + lhs/rhs


        :Parameters:
        ----------------
        other : dictable (or something that can be turned to one)
            what we exclude with.
        lcols : str/list of strs, optional
            the left columns/formulae on which we match. The default is None.
        rcols : str/list of strs, optional
            the right columns/formulae on which we match. The default is None.
        mode : string, optional
            When set to 'r', performs xor the other way. The default is 'l'.

        :Returns:
        -------
        dictable
            a dictable containing what is in self but not in ther other dictable.

        """
        if not isinstance(other, dictable):
            other = dictable(other)
        if lcols is None:
            lcols = self.keys() & other.keys()
        if rcols is None:
            rcols = lcols
        lcols = as_tuple(lcols); rcols = as_tuple(rcols)
        if len(lcols)!=len(rcols):
            raise ValueError('cannot xor-join when cols on either side mismatch in length %s vs %s'%(lcols, rcols))
        if len(lcols) == 0:
            return self.copy()
        lxs, lids = self._listby(lcols)
        rxs, rids = other._listby(rcols)
        mode = 1 if (is_str(mode) and mode[0].lower() == 'r') or mode == 1 else 0
        ls = len(lxs)
        rs = len(rxs)
        l = 0
        r = 0
        res = []
        while l<ls and r<rs:
            while l<ls and r<rs and cmp(lxs[l],rxs[r]) == -1:
                if mode == 0:
                    res.append(lids[l])
                l+=1
            while l<ls and r<rs and cmp(lxs[l],rxs[r]) == 1:
                if mode == 1:
                    res.append(rids[r])                    
                r+=1
            if l<ls and r<rs and lxs[l] == rxs[r]:
                r+=1
                l+=1
        if mode == 0:
            if l<ls:
                res.extend(lids[l:])
            return self[sum(res, [])]
        else:
            if r<rs:
                res.extend(rids[r:])
            return other[sum(res, [])]

    __mul__ = join
    __truediv__ = xor
    __div__ = xor
    
    def xyz(self, x, y, z, agg = None):
        """
        
        pivot table functionality.
        
        :Parameters:
        ----------------
        x : str/list of str
            unique keys per each row
        y : str
            unique key per each column
        z : str/callable
            A column in the table or an evaluated quantity per each row
        agg : None/callable or list of callables, optional
            Each (x,y) cell can potentially contain multiple z values. so if agg = None, a list is returned
            If you want the data aggregated in any way, then supply an aggregating function(s)

        :Returns:
        -------
        A dictable which is a pivot table of the original data
        

        :Example:
        -------
        >>> from pyg import *
        >>> timetable_as_list = dictable(x = [1,2,3]) * dictable(y = [1,2,3]) 
        >>> timetable = timetable_as_list.xyz('x','y',lambda x, y: x * y)
        >>> assert timetable = dictable(x = [1,2,3], )

        :Example:
        -------
        >>> self = dictable(x = [1,2,3]) * dictable(y = [1,2,3]) 
        >>> x = 'x'; y = 'y'; z = lambda x, y: x * y
        >>> self.exc(lambda x, y: x+y==5).xyz(x,y,z, len)
        
        """
        if not is_strs(x):
            raise ValueError('x must be columns %s'%x)
        agg = as_list(agg)
        x = as_tuple(x)
        xykeys = x + as_tuple(y)
        xys, ids = self._listby(xykeys)
        zs = self[z]
        y_ = y if is_str(y) else '_columns'
        rs = type(self)(xys, x + (y_,))        
        ys = rs[as_list(y_)].listby(y_)
        y2id = dict(zip(ys[y_], range(len(ys))))
        xs, yids = rs._listby(x)
        res = [[None for _ in range(len(ys))] for _ in range(len(xs))]
        for i in range(len(xs)):
            for j in yids[i]:
                xy = xys[j]
                k = y2id[xy[-1]]
                value = [zs[id_] for id_ in ids[j]]
                if agg:
                    for a in agg:
                        value = a(value)
                res[i][k] = value
        dx = type(self)(xs, x)
        dy = type(self)(res, y2id.keys())
        dx.update(dy)
        return dx
    
    pivot = xyz
    
    def unpivot(self, x, y, z):
        """
        undoes self.xyz / self.pivot
        
        :Example:
        -------
        >>> from pyg import *
        >>> orig = (dictable(x = [1,2,3,4]) * dict(y = [1,2,3,4,5]))(z = lambda x, y: x*y)
        >>> pivot = orig.xyz('x', 'y', 'z', last)
        >>> unpivot = pivot.unpivot('x','y','z').do(int, 'y') # the conversion to column names mean y is now string... so we convert back to int
        >>> assert orig == unpivot

        :Parameters:
        ----------------
        x : str/list of strings
            list of keys in the pivot table.
        y : str
            name of the columns that wil be used for the values that are currently column headers.
        z : str
            name of the column that describes the data currently within the pivot table.

        :Returns:
        -------
        dictable

        """
        xcols = as_list(x)
        if isinstance(y, dict) and len(y) == 1:
            y, ycols = list(y.items())[0]
        else:
            ycols = self.keys() - x
        ycols = as_tuple(ycols)
        n = len(ycols)

        res = type(self)({k: sum([[row[k]]*n for row in self], []) for k in xcols})
        res[y] = ycols * len(self)
        res[z] = sum([[row[ycol] for ycol in ycols] for row in self], [])
        return res
    

            

def is_dictable(value):
    return isinstance(value, dictable)




import pandas as pd
from pyg_base._dates import dt
from pyg_base._types import is_date, is_ts, is_str, is_num
from pyg_base._pandas import df_reindex

def years_between(t0, t1):
    """
    whole years between two dates

    Example
    -------
    >>> t0 = dt(2000,1,1)
    >>> t1 = dt(2001,1,1)
    >>> assert years_between(t0,t1) == 1
    >>> assert years_between(dt(2000, 2, 1), t1) == 0

    
    """
    y = t1.year - t0.year
    if dt(t0.year + y, t0.month, t0.day) > t1:
        return y-1
    else:
        return y

_tenors = {'spot': 0, 'sn' : 1/365, 'tn' : 2/365, 's/n': 1/365, 't/n': 2/365}

def years_to_maturity(maturity, ts = None):
    """
    calculates years to maturity with part of the year calculated as ACT/365
    
    :Example:
    ---------
    >>> from pyg import *     
    >>> ts = pd.Series(range(1000), drange(-999))
    >>> maturity = dt('2Y')
    >>> years_to_maturity(maturity, ts)
    
    :Example: ts is None
    --------------------
    >>> assert years_to_maturity('1y') == 1
    >>> assert years_to_maturity('2Q') == 0.5
    >>> assert years_to_maturity('18m') == 1.5
    >>> assert years_to_maturity('13w') == 0.25
    >>> assert years_to_maturity('730d') == 2
    >>> assert years_to_maturity('spot') == 0    
    """
    if isinstance(maturity, (list, tuple)):
        return type(maturity)([years_to_maturity(m, ts=ts) for m in maturity])
    elif is_str(maturity): ## years_to_maturity('5y',..)
        maturity = maturity.lower()
        if maturity in _tenors:
            return _tenors[maturity]
        else:
            return int(maturity[:-1]) / dict(q = 4, y = 1, m = 12, w = 52, d = 365, b = 252)[maturity[-1]]
    elif is_num(maturity):
        return maturity
    elif is_date(maturity):
        if ts is None:
            ts = dt(0)
        if is_date(ts):
            y = years_between(ts, maturity)
            frac = (dt(maturity, f'-{y}y') - ts).days / 365.
            return y + frac
        elif is_ts(ts):
            if len(ts) == 0:
                return ts
            t0 = ts.index[0]
            years = list(range(2+maturity.year-t0.year))[::-1]
            dates = [dt(maturity, f'-{y}y') for y in years]
            y = df_reindex(pd.Series(years, dates), ts, method = 'bfill')
            days = df_reindex(pd.Series(dates, dates), ts, method = 'bfill')
            frac = pd.Series((days.values - days.index).days / 365, ts.index)
            return y + frac
        elif isinstance(ts, list):
            return [years_to_maturity(maturity, t) for t in ts]             
        else:
            raise ValueError(f'cannot calculate years_to_maturity for {maturity} and {ts}')
    else:
        raise ValueError('not sure how to calculate years to maturity')


from copy import copy
from pyg_base._dictable import dictable
from pyg_base._as_list import as_list
from pyg_base._dict import _tree_setitem, dictattr, _tree_types


def _table_to_tree(tree, pattern, d, base, ignore = None, types = None):
    path = pattern.split('/')
    item = [d[p[1:]] if p.startswith('%') else p for p in path]
    _tree_setitem(tree, item, base = base, ignore = ignore, types = types)
    

def table_to_tree(tree, pattern, table, base = dictattr, ignore = None, types = None):
    """
    This is the reverse of tree_to_table and allows us to modify a tree using a table of values.
    
    We set an item that looks like this:

    >>> pattern = 'markets/%market/weight/%weight' 
    
    with a value that looks like this:
        
    >>> table = dictable(market = ['TY','ES'], weight = [0.3,0.7])
    
    >>> table_to_tree(None, pattern, table = table)
    
    {'markets': {'TY': {'weight': 0.3}, 'ES': {'weight': 0.7}}}
    
    """
    types = _tree_types(types)
    ignore = as_list(ignore)
    base = type(tree) if base is None else base
    tree = base() if tree is None else copy(tree)
    if isinstance(table, (dictable, list)):
        for row in table:
            _table_to_tree(tree, pattern, row, base = base, ignore = ignore, types = types)
    else:
        _table_to_tree(tree, pattern, table, base = base, ignore = ignore, types = types)
    return tree
    





import scipy
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from pyg_base._tenor import years_to_maturity
from pyg_base._types import is_ts, is_num, is_arr, is_df, is_strs, is_nan, is_date
from pyg_base._pandas import df_reindex
from pyg_base._loop import pd2np

def _val(x, i, n):
    if is_num(x):
        return x
    if len(x.shape) < 2:
        return x
    elif len(x.shape) == 2:
        return x[i]
    

@pd2np
def _interpolate(y,x,xnew, 
                   kind = 'linear', 
                   axis = -1, 
                   copy = False, 
                   bounds_error = False,
                   fill_value = np.nan,
                   assume_sorted = True,
                   min_n = 2):
    """
    
    :Example: simple interplation with nan
    ---------

    >>> x = np.array([1,2,3,4])
    >>> y = np.array([1,np.nan, np.nan, 2.])
    >>> assert _interpolate(x = x, y = y, xnew = -2, fill_value = 'extrapolate') == 0

    :Example: simple interplation with y in 2d
    ---------
    >>> x = np.array([1,2,3,4])
    >>> y = np.array([[1,np.nan, np.nan, 2.],
                      [4,5,6,7.]])
    >>> assert eq(_interpolate(y, x, 2.5), np.array([1.5, 5.5]))
    >>> assert eq(_interpolate(y, x, [2.5,2.5]), np.array([1.5, 5.5]))
    >>> assert eq(_interpolate(y, np.array([x,x]), [2.5,2.5]), np.array([1.5, 5.5]))

    """

    if len(y.shape) == 2:
        if is_arr(x) and len(x.shape)==2:
            xs = x
        else:
            xs = [x] * y.shape[0]
        if isinstance(xnew, list):
            xnew = np.array(xnew)
        if is_arr(xnew) and len(xnew.shape) == 2:
            xnews = xnew
        elif is_arr(xnew) and len(xnew.shape) == 1 and len(xnew) == y.shape[0]:
            xnews = xnew
        else:
            xnews = [xnew] * y.shape[0]
        res = np.array([_interpolate(x = xs[i], y = y[i], xnew = xnews[i], 
                           kind = kind, 
                           axis = axis, 
                           copy = copy, 
                           bounds_error = bounds_error,
                           fill_value = fill_value,
                           assume_sorted = assume_sorted)
                           for i in range(y.shape[0])])
        return res
    if is_nan(xnew):
        return np.nan
    mask = ~pd.isnull(y)
    x = x[mask]
    y = y[mask]
    if len(y) < min_n:
        return np.nan + xnew
    return interp1d(x, y, kind = kind, axis = axis, copy = copy, 
                    bounds_error = bounds_error, 
                    fill_value = fill_value,
                    assume_sorted = True)(xnew)

def interpolate(a, y, x = None, 
                   kind = 'linear', 
                   axis = -1, 
                   copy = False, 
                   bounds_error = False,
                   fill_value = np.nan,
                   assume_sorted = True,
                   xmethod = 'linear'):
    """
    implements scipy.interpolate.interp1d with support:
        - handling nan for y values 
        - pandas objects
    
    :Example:
    ---------
    >>> y = pd.DataFrame(np.array([[1,np.nan, np.nan, 2.],
                      [4,5,6,7.]]), index = drange(-1))
    
    >>> x = [0,1,2,3]    
    >>> xnew = 1.5
    >>> assert eq(interpolate(x,y,xnew), pd.Series([1.5, 5.5], drange(-1)))

    :Example: interpolated x
    ---------
    >>> y = pd.DataFrame(np.array([
                        [1,np.nan, np.nan, 2.],
                       [4,5,6,7.],
                       [7,8,9,10]]), index = drange(-2))
    
    >>> x = pd.DataFrame(np.array([
                       [4,5,6,7.],
                       [7,8,9,10]]), index = [dt(-2),dt(0)])

    >>> xnew = 7
    >>> assert eq(interpolate(x,y,xnew), pd.Series([2., 5.5, 7.], drange(-2)))
    >>> assert eq(interpolate(x,y,5, fill_value = 'extrapolate'), pd.Series([4/3., 3.5, 5.], drange(-2)))
    
    """
    if x is None:
        x = y.columns.values
    elif isinstance(x, list):
        x = np.array(x)
    if is_date(a) and is_ts(y):
        a = years_to_maturity(a, y)
    if isinstance(a, list):
        a = np.array(a)
    if is_ts(a):
        y = df_reindex(y, a) if is_ts(y) else y
        x = df_reindex(x, a, method = xmethod) if is_ts(x) else x
    if is_df(y):
        x = df_reindex(x, y, method = xmethod) if is_df(x) else x
    res = _interpolate(y = y,x = x,xnew = a, kind = kind, 
                           axis = axis, 
                           copy = copy, 
                           bounds_error = bounds_error,
                           fill_value = fill_value,
                           assume_sorted = assume_sorted)
    if is_df(res) and is_df(a) and res.shape[1] == a.shape[1]:
        res.columns = a.columns
    return res



from pyg_base._dates import dt, DAY, TMIN, TMAX, dt_bump, is_period, is_bump, ymd, period
from pyg_base._types import is_int, is_str, is_ts, is_arr, is_pd, is_nan
from pyg_base._as_list import as_list 
from pyg_base._dict import Dict
from pyg_base._cache import cache
from pyg_base._cfg import cfg_read

import datetime
from dateutil.rrule import rrule, MONTHLY, WEEKLY, DAILY, YEARLY, HOURLY, MINUTELY, SECONDLY, MO, TU, WE, TH, FR, SA, SU

import numpy as np
import re

weekdays = {0: MO, 1: TU, 2: WE, 3: TH, 4: FR, 5: SA, 6: SU}

HOUR = datetime.timedelta(hours = 1)
minute = datetime.timedelta(minutes = 1)
second = datetime.timedelta(seconds = 1)
microsecond = datetime.timedelta(microseconds = 1)
isoformat = re.compile('^[0-9]{4}-[0-9]{2}-[0-9]{2}T')

    
__all__ = ['date_range', 'drange', 'calendar', 'Calendar', 'clock', 'as_time']

cfg = cfg_read()

def _day_start(day_start = None):
    if day_start is None:
        res = cfg.get('day_start',0)
    else:
        res = day_start
    return as_time(res)

def _day_end(day_end = None):
    if day_end is None:
        res = cfg.get('day_end',235959)
    else:
        res = day_end
    return as_time(res)

def as_time(t = None):
    """
    parses t into a datetime.time object
    
    :Example:
    ---------
    >>> assert as_time('10:30:40') == datetime.time(10, 30, 40)
    >>> assert as_time('103040') == datetime.time(10, 30, 40)
    >>> assert as_time('10:30') == datetime.time(10, 30)
    >>> assert as_time('1030') == datetime.time(10, 30)
    >>> assert as_time('05') == datetime.time(5)
    >>> assert as_time(103040) == datetime.time(10, 30, 40)
    >>> assert as_time(13040) == datetime.time(1, 30, 40)
    >>> assert as_time(130) == datetime.time(1, 30)
    >>> assert as_time(datetime.time(1, 30)) == datetime.time(1, 30)
    >>> assert as_time(datetime.datetime(2000, 1, 1, 1, 30)) == datetime.time(1, 30)
    
    :Parameters:
    ----------
    t : str/int/datetime.time/datetime.datetime
        time of day

    :Returns:
    -------
    datetime.time

    """
    if t is None:
        t = dt()
    if is_str(t):
        hms = t.split(':')
        if len(hms) == 3:
            h,m,s = hms
            return datetime.time(int(h), int(m), int(s))
        elif len(hms) == 2:
            h,m = hms
            return datetime.time(int(h), int(m))
        else:
            t = int(t)
    if is_int(t):
        if t<100:
            return datetime.time(t)
        elif t<10000:
            h = t // 100
            m = t % 100
            return datetime.time(h, m)
        elif t<1000000:
            h = t // 10000
            t = t - h * 10000
            m = t // 100
            s = t % 100
            return datetime.time(h,m,s)
    elif isinstance(t, datetime.time):
        return t
    elif isinstance(t, datetime.datetime):
        return datetime.time(t.hour, t.minute, t.second)
    else:
        raise ValueError('cannot parse time %s'%t)

_LY = dict(b = DAILY, d = DAILY, w = WEEKLY, m = MONTHLY, q = MONTHLY, y = YEARLY, h = HOURLY, n = MINUTELY, s = SECONDLY)

def drange(t0 = None, t1 = None, bump = None):
    """
    A quick and happy wrapper for dateutil.rrule

    :Examples:
    --------------
    >>> drange(2000, 10, 1) # 10 days starting from dt(2000,1,1)
    >>> drange(2000, '10b', '1b') # weekdays between dt(2000,1,1) and dt(2000,1,17)
    >>> drange('-10b', 0, '1b') # business days since 10 bdays ago
    >>> drange('-10b', '10b', '1w') # starting 10b days ago, to 10b from now, counting in weekly jumps
    
    :Parameters:
    ----------------
    t0 : date, optional
        start date. The default is None.
    t1 : date, optional
        end date. The default is None.
    bump : timedelta, int, string, optional
        bump period. The default is None.

    :Returns:
    -------
    list of dates
    
    :Example:
    ---------
    >>> t0 = 2000; t1 = 1999
    >>> bump = '-1b'
    
    :Example:
    ---------
    >>> t0 = dt(2020); t1 = dt(2021); bump = datetime.timedelta(hours = 4)

    """
    t0, t1 = date_range(t0, t1)  
    if t0 == t1:
        return [t0]        
    if bump is None:
        bump = 1 if t0<t1 else -1
    if is_int(bump):
        if (t1-t0).days * bump <= 0:
            raise ValueError('cannot go from %s to %s in steps of %s'%(t0,t1,bump))
        freq = DAILY
        res = list(rrule(freq, interval = 1, dtstart = min(t0,t1), until = max(t0,t1))) # unfortunately does not actually work for negative bumps
        res = res[::-1] if bump<0 else res
        res = res[::abs(bump)] if abs(bump)>1 else res
        return res
    elif isinstance(bump, datetime.timedelta):
        t = t0
        res = []
        if t1>t0: 
            if t0 + bump <= t0:
                raise ValueError('cannot move forward from %s to %s using %s'%(t0, t1, bump))
            while t<=t1:
                res.append(t)
                t = t + bump
            return res
        elif t1<t0: 
            if t0 + bump >= t0:
                raise ValueError('cannot move back from %s to %s using %s'%(t0, t1, bump))
            while t>=t1:
                res.append(t)
                t = t + bump
            return res
        else:
            return [t0]
    elif is_period(bump):
        bump = bump.lower()
        bmp = period.search(bump).group()
        if bump == bmp: ## single bump
            prd = bump[-1]
            interval = int(bump[:-1]) * dict(q = 3).get(prd ,1)
            if (t1-t0).days * interval < 0:
                raise ValueError('cannot go from %s to %s in steps of %s'%(t0,t1,bump))
            freq = _LY[prd]
            if prd == 'b':
                res = [t for t in rrule(freq, interval = 1, dtstart = min(t0,t1), until = max(t0,t1)) if t.weekday()<5]
                res = res[::-1] if interval<0 else res    
                res = res[::abs(interval)] if abs(interval)>1 else res
                return res            
            else:
                return list(rrule(freq, interval = interval, dtstart = t0, until = t1))
        else:
            t = t0
            res = []
            if t1>t0: 
                if dt_bump(t0, bump) <= t0:
                    raise ValueError('cannot move forward from %s to %s using %s'%(t0, t1, bump))
                while t<=t1:
                    res.append(t)
                    t = dt_bump(t, bump)
                return res
            elif t1<t0: 
                if dt_bump(t0, bump) >= t0:
                    raise ValueError('cannot move back from %s to %s using %s'%(t0, t1, bump))
                while t>=t1:
                    res.append(t)
                    t = dt_bump(t, bump)
                return res
            else:
                return [t0]
                            

class _calendar():
    def dt_bump(self, t, bump):
        return dt_bump(t, bump)
    
    def date_range(self, t0 = None, t1 = None):
        """
        creates a date range boundaries from start/end points
    
        :Parameters:
        ----------------
        t0 : date or a date bump, optional
            start date. The default is None.
        t1 : date or a date bump, optional
            end date. The default is None.
    
        :Returns:
        -------
        sorted list of two dates
        
        
        :Examples:
        --------------
        >>> t = dt.today()
        >>> assert date_range() == [dt(1900), t]
        >>> assert date_range(-100, 100) == [dt_bump(t, -100), dt_bump(t,100)]
        >>> assert date_range(-100, '10b') == [dt_bump(t, -100), dt_bump(t,'10b')]
        >>> assert date_range(2000, '10b') == [dt(2000,1,1), dt(2000,1,17)]
    
        """
        t = dt(0)
        if t1 is None:
            if t0 is None:
                return [TMIN, t]
            elif is_bump(t0):
                return sorted([t, self.dt_bump(t, t0)])
            else:
                return sorted([t, dt(t0)])
        elif is_bump(t1):
            if t0 is None:
                return [t, self.dt_bump(t, t1)]
            elif is_bump(t0):
                return [self.dt_bump(t, t0), dt_bump(t, t1)]
            else:
                t0 = dt(t0)
                return [t0, self.dt_bump(t0, t1)]
        else:
            t1 = dt(t1)
            if t0 is None:
                return [t, t1]
            elif is_bump(t0):
                return [self.dt_bump(t1, t0), t1]
            else:
                return [dt(t0), t1]


_c  = _calendar()

def date_range(t0 = None, t1 = None):
    return _c.date_range(t0, t1)
        

"""
calendars is implemented as a factory of calendars indexed by key

We therefore choose to make any modifications to the calendar object in-place
"""
calendars = dict()

class Calendar(Dict, _calendar):
    """
    Calendar is 
        - a dict 
        - containing holiday dates 
        - implementing business day arithmetic

    Calendar is restricted to operate between cal.t0 and cal.t1 which default to TMIN = 1900 and TMAX = 2300

    Calendar does this by having two key members:
        - dt2int: a mapping from all business dates to their integer 'clock'
        - int2dt: a mapping from integer value to the date
    
    Since Calendar is an 'expensive' memory wise, we assign a key to the calendar and the Calendar is stored in the singleton calendars under this key
    
    :Example:
    -------
    >>> from pyg import *
    >>> holidays = dictable([[1,'2012-01-02','New Year Day',],
                            [2,'2012-01-16','Martin Luther King Jr. Day',],
                            [3,'2012-02-20','Presidents Day (Washingtons Birthday)',],
                            [4,'2012-05-28','Memorial Day',],
                            [5,'2012-07-04','Independence Day',],
                            [6,'2012-09-03','Labor Day',],
                            [7,'2012-10-08','Columbus Day',],
                            [8,'2012-11-12','Veterans Day',],
                            [9,'2012-11-22','Thanksgiving Day',],
                            [10,'2012-12-25','Christmas Day',],
                            [11,'2013-01-01','New Year Day',],
                            [12,'2013-01-21','Martin Luther King Jr. Day',],
                            [13,'2013-02-18','Presidents Day (Washingtons Birthday)',],
                            [14,'2013-05-27','Memorial Day',],
                            [15,'2013-07-04','Independence Day',],
                            [16,'2013-09-02','Labor Day',],
                            [17,'2013-10-14','Columbus Day',],
                            [18,'2013-11-11','Veterans Day',],
                            [19,'2013-11-28','Thanksgiving Day',],
                            [20,'2013-12-25','Christmas Day',],
                            [21,'2014-01-01','New Year Day',],
                            [22,'2014-01-20','Martin Luther King Jr. Day',],
                            [23,'2014-02-17','Presidents Day (Washingtons Birthday)',],
                            [24,'2014-05-26','Memorial Day',],
                            [25,'2014-07-04','Independence Day',],
                            [26,'2014-09-01','Labor Day',],
                            [27,'2014-10-13','Columbus Day',],
                            [28,'2014-11-11','Veterans Day',],
                            [29,'2014-11-27','Thanksgiving Day',],], ['i', 'date', 'name']).do(dt, 'date')

    
    >>> cal = calendar('US', holidays.date, t0 = 2012, t1 = 2015)
    >>> assert not cal.is_bday(dt(2013,9,2))        # Labor day

    >>> cached_calendar = calendar('US')
    >>> assert not cached_calendar.is_bday(dt(2013,9,2))   # Labor day

    >>> assert cal.adjust(dt(2013,9,2)) == dt(2013,9,3)
    >>> assert cal.drange(dt(2013,9,0), dt(2013,9,7), '1b') == [dt(2013,8,30), dt(2013,9,3), dt(2013,9,4), dt(2013,9,5), dt(2013,9,6),] ## skipped labour day and weekend prior
    
    >>> assert cal.bdays(dt(2013,9,0), dt(2013,9,7)) == 5
    """
    def __init__(self, key = None, holidays = None, weekend = None, t0 = None, t1 = None, adj = 'm'):
        """
        :Parameters:
        ----------------
        key : str, None, optional
            The key used to cache the calendar in singleton. The default is None.
        holidays : list of dates
            list of holidays
        weekend : list of int, optional
            What days are weekend 0=MON... 6 = SUN. The default is [5,6].
        t0 : datetime, optional
            calendar's first date. The default is TMIN = 1900.
        t1 : datetime, optional
            calendar's last date. The default is TMIN = 2300.
        adj : char, optional
            how non-bdays are adjusted. The default is 'm'.
            Three conventions are used:
                'f'ollowing : next date
                'p'revious: previous date
                'm'odified following: following date unless following date is in a different month, when we use previous.

        :Returns:
        -------
        Calendar object.

        """
        if isinstance(key, dict):
            super(Calendar, self).__init__(**key)
        if isinstance(key, Calendar):
            super(Calendar, self).__init__(key)
        else:
            if t1 is None:
                t1 = TMAX
            if t0 is None:
                t0 = TMIN
            t0, t1 = date_range(t0, t1)
            ## weekend are weekend days
            if weekend is None: 
                weekend = [5,6]
            else:
                weekend = as_list(weekend)
            holidays = as_list(holidays)
            holidays = dict(zip(holidays, holidays)) # we prefer to store holidays as a dict, as check of date in holidays is faster for hash
            super(Calendar, self).__init__(weekend = weekend, holidays = holidays,
                                           key = key, t0 = t0, t1 = t1, adj = adj)


    def _populate(self):
        """
        This is run once as it is quite expensive
        """
        if self.get('dt2int') is None or self.get('int2dt') is None:
            byweekday = tuple(v for k,v in weekdays.items() if k not in self.weekend)
            bdays = [date for date in rrule(DAILY, interval = 1, dtstart = self.t0, until = self.t1, byweekday = byweekday) if date not in self.holidays]
            self['dt2int'] = dict(zip(bdays, range(len(bdays))))        
            self['int2dt'] = dict(zip(range(len(bdays)), bdays))
        return self
        
    
    def __repr__(self):
        return 'calendar(%(key)s) from %(t0)s to %(t1)s'%self
    
    def is_holiday(self, date):
        return date.weekday() in self.weekend or ymd(date) in self.holidays
    
    def is_bday(self, date):
        return date.weekday() not in self.weekend and ymd(date) not in self.holidays

    def is_trading(self, date = None, day_start = None, day_end = None):
        """
        calculates if we are within a trading session 

        :Parameters:
        ----------
        date : datetime, optional
            the time & date we want to check. The default is None (i.e. now)

        :Returns:
        -------
        bool:
            are we within a trading session
        """
        tod = as_time(date)
        day_start = _day_start(day_start)
        day_end = _day_end(day_end)
        day = ymd(date)
        if day_end >= day_start:
            return tod <= day_end and tod >= day_start and self.is_bday(day)
        else:
            if tod>=day_start and self.is_bday(day + DAY): # we are within tomorrow's session
                return True
            elif tod<=day_end and self.is_bday(day): # we are within today's session
                return True
            else:
                return False

    def trade_date(self, date = None, adj = None, day_start = None, day_end = None):
        """
        This is very similar for adjust, but it also takes into account the time of the day.
        if day_start = 0 and day_end = 23:59:59 then this is exactly adjust.

        :Parameters:
        ----------
        date : datetime, optional
            date (with time). The default is None.
        adj : f/p, optional
            If date isn't within trading day, which direction to adjust to? The default is None.
            
        
        :Example:
        ---------
        >>> from pyg import *; import datetime

        >>> uk = calendar('UK', day_start = 8, day_end = 17)
        >>> assert uk.trade_date(dt(2021,2,9,5), 'f') == dt(2021, 2, 9)  # Tuesday morning rolls into Tuesday
        >>> assert uk.trade_date(dt(2021,2,9,5), 'p') == dt(2021, 2, 8)  # Tuesday morning back into Monday
        >>> assert uk.trade_date(dt(2021,2,7,5), 'f') == dt(2021, 2, 8)  # Sunday rolls into Monday
        >>> assert uk.trade_date(dt(2021,2,7,5), 'p') == dt(2021, 2, 5)  # Sunday rolls back to Friday

        >>> assert uk.trade_date(date = dt(2021,2,9,23), adj = 'f') == dt(2021, 2, 10)  # Tuesday eve rolls into Wed
        >>> assert uk.trade_date(date = dt(2021,2,9,23), adj = 'p') == dt(2021, 2, 9)  # Tuesday eve back into Tuesday
        >>> assert uk.trade_date(date = dt(2021,2,7,23), adj = 'f') == dt(2021, 2, 8)  # Sunday rolls into Monday
        >>> assert uk.trade_date(date = dt(2021,2,7,23), adj = 'p') == dt(2021, 2, 5)  # Sunday rolls back to Friday

        >>> assert uk.trade_date(date = dt(2021,2,9,12), adj = 'f') == dt(2021, 2, 9)  # Tuesday is Tuesday
        >>> assert uk.trade_date(date = dt(2021,2,9,12), adj = 'p') == dt(2021, 2, 9)  # Tuesday is Tuesday

        >>> au = calendar('AU', day_start = 2230, day_end = 1300)
        >>> assert au.trade_date(dt(2021,2,9,5), 'f') == dt(2021, 2, 9)  # Tuesday morning in session
        >>> assert au.trade_date(dt(2021,2,9,5), 'p') == dt(2021, 2, 9)  # Tuesday morning in session
        >>> assert au.trade_date(dt(2021,2,7,5), 'f') == dt(2021, 2, 8)  # Sunday rolls into Monday
        >>> assert au.trade_date(dt(2021,2,7,5), 'p') == dt(2021, 2, 5)  # Sunday rolls back to Friday

        >>> assert au.trade_date(date = dt(2021,2,9,23), adj = 'f') == dt(2021, 2, 10)  # Tuesday eve rolls into Wed
        >>> assert au.trade_date(date = dt(2021,2,9,23), adj = 'p') == dt(2021, 2, 10)  # Already in Wed
        >>> assert au.trade_date(date = dt(2021,2,7,23), adj = 'f') == dt(2021, 2, 8)  # Sunday rolls into Monday
        >>> assert au.trade_date(date = dt(2021,2,7,23), adj = 'p') == dt(2021, 2, 8)  # Already on Monday
        >>> assert au.trade_date(date = dt(2021,2,5,23), adj = 'f') == dt(2021, 2, 8)  # Friday afternoon rolls into Monday

        >>> assert au.trade_date(date = dt(2021,2,9,14), adj = 'f') == dt(2021, 2, 10)  # Tuesday is over, roll to Wed
        >>> assert au.trade_date(date = dt(2021,2,9,14), adj = 'p') == dt(2021, 2, 9)  # roll back to Tues
            
        """
        # first we split it into day and time of day
        tod = as_time(date)
        day = ymd(date)
        adj = (adj or self.adj)[0].lower()
        day_start = _day_start(day_start); day_end = _day_end(day_end)        
        if day_end >= day_start:
            res = self.adjust(day, adj)
            if tod > day_end: # post today trading session
                if adj == 'f':
                    if res > day:
                        return res
                    else:
                        return self.add(res, 1)
                elif adj == 'p':
                    return res
            elif tod < day_start: # pre today session
                if adj == 'f':
                    return res
                elif adj == 'p':
                    if res < day:
                        return res
                    else:
                        return self.add(res, -1)
            else: # in session. 
                return res
        else:
            if tod > day_start: # we are in tomorrow's trading session
                tomorrow = day + datetime.timedelta(1)
                res = self.adjust(tomorrow, adj)
                return res
            elif tod < day_end: # we are in today's trading session
                res = self.adjust(day , adj)
                return res
            else: # we are in post-session
                res = self.adjust(day, adj)
                if adj == 'f':
                    if res > day:
                        return res
                    else:
                        return self.add(res, 1)
                elif adj == 'p':
                    return res

    
    def adjust(self, date, adj = None):
        """

        adjust a non-business day to prev/following bussiness date

        :Parameters:
        ----------------
        date : datetime.
        adj : None or p/f/m
            adjustment convention: 'prev/following/modified following'

        :Returns:
        -------
        dateime
            nearby business day

        """
        adj = (adj or self.adj or 'm').lower()
        t = ymd(date)
        t = datetime.datetime(t.year, t.month, t.day)
        
        if adj.startswith('f'):
            while self.is_holiday(t) and t <= self.t1:
                t = t + DAY
            while t > self.t1 and t.weekday() in self.weekend:
                t = t + DAY
            return t
        elif adj.startswith('p'):
            while self.is_holiday(t) and t >= self.t0:
                t = t - DAY
            while t < self.t0 and t.weekday() in self.weekend:
                t = t - DAY
            return t
        elif adj.startswith('m'): #modified following
            t = self.adjust(date, 'f')
            if t.month!=date.month:
                return self.adjust(date, 'p')
            else:
                return t
        else:
            raise ValueError(f'date adjustment {adj} must start with f(ollowing), p(revious) or m(odified)')

    def dt_bump(self, t, bump, adj = None):
        """
        adds a bump to a date

        :Parameters:
        ----------------
        t : datetime
            date to bump.
        bump : int, str
            bump e.g. '-1y' or '1b' or 3
        adj : adjustement type
            The default is None.

        :Returns:
        -------
        datetime
            bumped date.

        """
        if is_str(bump):
            while period.search(bump) is not None:
                bmp = period.search(bump).group().lower()
                bump = bump[len(bmp):]
                if bmp.endswith('b'):
                    if bmp == '+0b':
                        t = self.adjust(t, adj = 'f')
                    elif bmp == '-0b':
                        t = self.adjust(t, adj = 'p')                        
                    t = self.add(t, int(bmp[:-1]), adj = adj)
                else:
                    t = dt_bump(t, bmp)
            return t
        else:
            return dt_bump(t, bump)
        
    def clock(self, date):
        self._populate()
        date = ymd(date)
        return self.dt2int.get(date, self.dt2int[self.adjust(date)])

    def add(self, date, days, adj = None, aggregate = 'last'):
        """
        adjustes the start date to a business day. Then add business days on top.
        add will initiate self._populate unless we are bumping date by exactly one day

        Parameters
        ----------
        date : datetime or timeseries
            start date.
        days : int
            days to bump.
        adj : str, optional
            adjusting method for date if it isn't a business day to start with.
        aggregate: str/function
            if adding days to a timeseries, we can get multiple starting dates adding to same date. By default, pick last value
        Returns
        -------
        datetime
            bumped date.
            
        Example: aggregating of timeseries data over holidays
        --------
        >>> from pyg import * 
        >>> date = pd.Series(range(10), drange(9)) ## goes over the weekend, so multiple dates will be bumped to Monday
        >>> cal = calendar()
        >>> assert len(cal.add(date, 1)) < len(date)
        """
        adj = adj or self.adj
        if is_ts(date):
            res = date.copy()
            res.index = [self.add(t, days, adj = adj) for t in res.index]
            if len(set(res.index)) < len(res):
                res.index.name = res.index.name or 'date'
                res = res.groupby(res.index.name).apply(aggregate)
            return res            
        t = self.adjust(date, adj)
        if abs(days)>1:
            self._populate()
            return self.int2dt[self.dt2int[t] + days]
        else:
            increment = days * DAY
            res = t + increment
            while self.is_holiday(res):
                res = res + increment
        return res
    
    def bdays(self, t0, t1, adj = None):
        self._populate()
        adj = adj or self.adj
        return self.dt2int[self.adjust(t1, adj)] - self.dt2int[self.adjust(t0, adj)]
    
    def drange(self, t0 = None, t1 = None, bump = None):
        t0, t1 = self.date_range(t0, t1)
        if is_str(bump) and bump[-1] == 'b':
            self._populate()
            b = int(bump[:-1])
            i0 = self.dt2int[self.adjust(t0)]
            i1 = self.dt2int[self.adjust(t1)]
            return [self.int2dt[i] for i in range(i0, i1+b, b)]
        else:
            return drange(t0, t1, bump)
    
    def filter(self, ts, day_start = None, day_end = None):
        if day_start is None or day_end is None:
            return ts[[self.is_bday(date) for date in ts.index]]
        else:
            day_start = _day_start(day_start)
            day_end = _day_end(day_end)
            return ts[[self.is_trading(date, day_start = day_start, day_end = day_end) for date in ts.index]]

        
def calendar(key = None, holidays = None, weekend = None, t0 = None, t1 = None):
    """
    A function to returns either an existing calendar or construct a new one.
    - calendar('US') will return a US calendar if that is already cached
    - calendar('US', us_holiday_dates) will construct a calendar with holiday dates and then cache it
    """
    if isinstance(key, Calendar):
        if holidays is None and weekend is None and t0 is None and t1 is None:
            calendars[key.key] = key
            key = key.key
        else:
            holidays = holidays or list(key.holidays.keys())
            weekend = weekend or key.weekend
            t0 = t0 or key.t0
            t1 = t1 or key.t1
            key = key.key
            calendars[key] = Calendar(key, holidays = holidays, weekend = weekend, t0 = t0, t1 = t1)            
    elif key not in calendars or holidays is not None or weekend is not None or t0 is not None or t1 is not None:
        calendars[key] = Calendar(key, holidays = holidays, weekend = weekend, t0 = t0, t1 = t1)
    return calendars[key]


@np.vectorize
def _weekday_intraday_clock(t):
    t = dt(t)
    j = ymd(t)
    c = calendar()._populate()
    if j in c.dt2int:
        i = t - j
        return c.dt2int[j] + (i.microseconds / 1e6 + i.seconds) / 86400
    else:
        j = c.adjust(j)        
        return c.dt2int[j]

@np.vectorize
def _fraction_clock(t):
    t = dt(t)
    return t.toordinal() + t.hour / 24 + t.minute / 1440 + (t.second + t.microsecond/1e6) / 86400

@np.vectorize
def _day_clock(t):
    t = ymd(t)
    return t.toordinal() 

@np.vectorize
def _week_clock(t):
    t = ymd(t)
    return t.toordinal() // 7

@np.vectorize
def _month_clock(t):
    t = ymd(t)
    return t.year*12 + t.month

@np.vectorize
def _year_clock(t):
    t = ymd(t)
    return t.year

@np.vectorize
def _quarter_clock(t):
    """
    >>> for m, q in zip(range(1,13), [0,0,0,1,1,1,2,2,2,3,3,3]):
    >>>     assert _quarter_clock(dt(2000,m,1)) == 8000 + q
    """
    t = dt(t)
    return t.year * 4 + (t.month-1) // 3

_clocks = dict(f = _fraction_clock, 
               d = _day_clock,
               w = _week_clock,
               m = _month_clock,
               y = _year_clock,
               q = _quarter_clock,
               k = _weekday_intraday_clock
               )


def clock(ts, time = None, t = None):
    """
    returns a vector marking the passage of time.

    :Parameters:
    ----------------
    ts : timeseries
    time : None, a string or a Calendar, or already a timeseries of times
        None: Will increment by 1 every non-nan observation
        'i' : increment by 1 every date in index (nan or not)
        'b' : weekdays distance
        'd' : day-distance (ignore intraday stamp)
        'f' : fraction-of-day-distance (do not ignore intraday stamp)
        'm' : month-distance
        'q' : quarter-distance
        'y' : year-distance
        calendar: uses the business-days distance between any two dates
        
    t: starting time in the past. 
    
    :Returns:
    -------
    an array 
        an increasing array of time such that distance between points match the above.
        
        
    :Example:
    -------
    >>> from pyg import *
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, 9))), np.arange(1,11))
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, 9)), t = 5), np.arange(6,16))
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, 9)), 'i'), np.arange(1,11))
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, 9)), 'b'), np.array([26090, 26090, 26090, 26091, 26092, 26093, 26094, 26095, 26095, 26095]))
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, '9b', '1b')), 'b'), np.arange(26090, 26100))

    >>> assert eq(clock(np.arange(10)), np.arange(1,11))
    >>> assert eq(clock(pd.Series(np.arange(10)), t = 5), np.arange(6,16))
    >>> assert eq(clock(np.arange(10), 'i'), np.arange(1,11))

    """
    if t is None or is_nan(t):
        t = 0
    if is_str(time):
        time = time[0].lower()
    if is_ts(time) or is_arr(time):
        return time
    elif time is None and (is_arr(ts) or is_pd(ts)):
        return np.full(len(ts), np.nan)
        # mask = np.min(np.isnan(ts), axis = 1) if len(ts.shape) == 2 else np.isnan(ts)
        # res = np.full(len(ts), 1)
        # res[mask] = 0
        # return t + np.cumsum(res)        
    elif time == 'i' and (is_arr(ts) or is_pd(ts)): #index
        return t + np.arange(1,1+len(ts))
    elif is_ts(ts):    
        if isinstance(time, Calendar):
            return np.vectorize(time.clock)(ts.index)
        elif callable(time):
            return np.vectorize(time)(ts.index)            
        elif time == 'b':
            time = calendar()
            return np.vectorize(time.clock)(ts.index)
        elif time in _clocks:
            return _clocks[time](ts.index)
        else:
            raise ValueError('cannot determine time for ts=%s, time=%s'%(ts, time))            
    else:
        raise ValueError('cannot determine time for ts=%s, time=%s'%(ts, time))




from pyg_base._dictable import dictable
from pyg_base._as_list import as_list
from pyg_base._dates import dt
from pyg_base._pandas import df_slice, as_series
from pyg_base._dictattr import dictattr
import pandas as pd

_data = 'data'

def _min(*values):
    values = [v for v in values if v is not None]
    if len(values) == 0:
        return None
    else:
        return min(values)

def _max(*values):
    values = [v for v in values if v is not None]
    if len(values) == 0:
        return None
    else:
        return max(values)

def df_roll_off(chain, loader,load_on = None,transform = None, live_check = None,
            roll = 'roll', expiry = '-1m', n = 0, data = None, cutoff = '-2b', do_if_no_n = False):
    """
    
    Creates a single timeseries from multiple timeseries data, 
    We assume we roll off a timeseries when it expires

    Parameters
    ----------
    chain : dictable
        list of timeseries to be loaded
    loader : callable
        function to load timeseries
    load_on : str/list of str
        columns to use when loading. The default is None.
    transform : callable, optional
        a transform to apply to the data. The default is None.
    live_check: callable, optional
        for data that is considered "live", i.e. it is data post cutoff date,
        you can apply additional transforms that are usually associated with checks that data is available
    roll: str, optional
        the column used to determine roll. The default is 'roll'.
    expiry : str/int, optional
        Relative date to today that determines the expiry of available timeseries, The default is '-1b'.
    n : integer, optional
        The size of the resulting curve. The default is 0, which corresponds to a SERIES of size 1
    data : timeseries, optional
        The rolled timeseries, previously calculated . The default is None.
    cutoff : str/int, optional
        old data will be used up to that relative date
    do_if_no_n: bool/callable
        We want to ensure that we have data available post the cutoff. 
        We need n timeseries. if we have less, we can choose to raise, or to call a function

    Returns
    -------
    timeseries. A rolled timeseries
    
    Example: roll when each time series ends
    --------
    >>> from pyg import * 
    >>> def loader(end_date, value):
    >>>     return pd.DataFrame({0:value}, drange(-800, min(dt(0), dt(end_date))))

    >>> end_dates = drange(dt(2000,4,0), dt(1000), '3m')
    >>> chain = dictable(value = range(len(end_dates)), end_date = end_dates)
    >>> load_on = None; transform = None; roll = 'roll'; expiry = '-1m'; n = 1; data = None; cutoff = '-1b'

    >>> d = df_roll_off(chain, loader = loader, n = 0)

    >>> d.chain ## we roll when data ends
    dictable[51 x 3]
    value|end_date           |roll               
    0    |2000-03-31 00:00:00|2000-03-31 00:00:00
    1    |2000-12-31 00:00:00|2000-12-31 00:00:00
    2    |2001-03-31 00:00:00|2001-03-31 00:00:00
    ...51 rows...
    48   |2024-03-31 00:00:00|None               
    49   |2024-12-31 00:00:00|None               
    50   |2025-03-31 00:00:00|None               

    >>> d.data[diff(d.data) > 0]
    2000-04-01     1
    2001-01-01     2
    ...
    2022-01-01    44
    2022-04-01    45
    
    
    Example: extend an existing result
    -----------------------------------
    >>> d2 = df_roll_off(chain, loader = loader, data = d.data.iloc[-100:])
    >>> assert eq(pd.DataFrame(d.data), pd.DataFrame(d2.data))
    
    Example: roll dates specified
    -----------------------------
    >>> chain = chain(roll = lambda end_date: dt(end_date, -15)) - 'data' ## roll mid-month
    >>> d3 = df_roll_off(chain, loader = loader).data
    >>> d3[diff(d3)>0]
    Out[163]: 
    2000-03-17     1 # <--- mid month rolls
    2000-12-17     2
                  ..
    2021-12-17    44
    2022-03-17    45
    Length: 45, dtype: int64
    

    >>> d3 = df_roll_off(chain, loader = loader, data = d2.data[:dt(2010)]).data 
    >>> d3[diff(d3)>0]
    Out[165]: 
    2000-04-01     1 ## <--- historic data that rolled end-of-month
    2001-01-01     2
                  ..
    2021-12-17    44 ## <--- data from 2010 that rolls mid-month
    2022-03-17    45
    Length: 45, dtype: int64
    

    Example: larger curve size
    --------------------------
    >>> dd = df_roll_off(chain, loader = loader, n = 4)
    >>> dd.data[dt(2000):dt(0)]

                   0     1     2     3
    2000-01-01   0.0   1.0   2.0   3.0
    2000-01-02   0.0   1.0   2.0   3.0
             ...   ...   ...   ...
    2022-09-05  45.0  46.0  47.0  48.0
    2022-09-06  45.0  46.0  47.0  48.0
    
    Example: cutoff behaviour
    --------
    >>> from pyg import *
    >>> factory = dict(a = pd.Series(range(51), drange(-100,-50)),
                   b = pd.Series(range(71), drange(-70,0)))
    >>> chain = dictable(ticker = ['a', 'b'])
    >>> loader = timer(lambda ticker: factory[ticker])

    >>> ts = df_roll_off(chain = chain, loader = loader, n = 1).data
    >>> assert ts.index[-1] == dt(0)

    ## we now make 'a' go all the way, just past the cutoff
    >>> alive_just_about = drange(-100,'-1b')
    >>> factory['a'] = pd.Series(range(len(alive_just_about)), alive_just_about)
    >>> ts = df_roll_off(chain = chain, loader = loader, n = 1).data
    >>> assert ts.index[-1] == dt('-1b')

    """
    n_ = max(n, 1)
    chain = dictable(chain)
    if roll not in chain.keys():
        chain[roll] = None
    if _data not in chain:
        chain[_data] = None
    data_ok = data is not None and len(data) > 0

    if data_ok and n==0:
        if len(data.shape) == 2:
            data = as_series(data.iloc[:,0])
    if data_ok and n==1:
        if len(data.shape) == 2 and data.shape[1] > 1:
            data = data.iloc[:,0]
    if data_ok and n > 1:
        if len(data.shape) == 1 or data.shape[1] < n:
            data_ok = False
        elif data.shape[1] > n:
            data = data.iloc[:,:n]
    if cutoff is not None:
        cutoff = dt(cutoff)
    if data_ok:
        data_cutoff = data.index[-1]
        if cutoff < data_cutoff:
            data_cutoff = cutoff
            data = df_slice(data, ub = data_cutoff, openclose = '[]')
    else:
        data_cutoff = None

    if data_ok:
        old_data = data
        old_chain = chain[[row[roll] is not None and row[roll]<data_cutoff for row in chain]]
        chain = chain[[row[roll] is None or row[roll]>=data_cutoff for row in chain]]
    else:
        old_data = None
        old_chain = None

    load_on = as_list(load_on)
    size = len(chain)

    ## we load the data, loading only live data
    now = dt(0)
    loaded_data = []
    i = j = 0
    while j < size and i < n_:
        row = chain[j]
        roll_date = row[roll]
        if roll_date is not None and data_cutoff is not None and roll_date < data_cutoff : ## old data
            loaded = row[_data]
        else: ## we load the data
            if len(load_on) == 0:
                loaded = row[loader]
            else:
                loaded = loader(**row[load_on])
            if (roll_date is None or roll_date >= now) and loaded is not None and len(loaded) > 0 and (loaded.index[-1] >= (cutoff or now)):
                if live_check is not None:
                    for check in as_list(live_check):
                        loaded = check(loaded)
                i = i + 1
        loaded_data.append(loaded)
        j = j + 1
    if i < n_:
        if do_if_no_n is True:
            raise ValueError(f'there are only {i} dataseries available post cutoff and {n} are required')
        elif callable(do_if_no_n):
            return do_if_no_n(i, n_)

    if j < size:
        loaded_data.extend([None] * (size - j))
    chain[_data] = loaded_data

    ## we now determine missing rolling dates
    expiry = dt(expiry)
    rolls = chain[roll]
    lb = None
    for j in range(size-1, -1, -1):
        loaded = loaded_data[j]
        if rolls[j] is None and loaded is None:
            rolls[j] = lb
        elif loaded is not None and len(loaded) > 0:
            lb = loaded.index[0]
            if loaded.index[-1] < expiry and rolls[j] is None:
                rolls[j] = loaded.index[-1]
    chain[roll] = rolls

    ### we now begin the roll in earnest
    c = chain.inc(lambda data: data is not None and len(data) > 0)
    if len(c) == 0: # no new data
        return dictattr(chain = old_chain + chain, data = old_data, roll = roll)    
    ub = [_min(row[roll], row[_data].index[-1]) for row in c]
    if transform is not None:
        c = c.do(transform, _data)
    new_data = df_slice(c[_data], ub = ub, openclose = '(]', n = n_)
    if n == 0:
        new_data = as_series(new_data)
    if old_data is not None and len(old_data) > 0:
        new_data = df_slice(new_data, lb = old_data.index[-1], openclose = '(]')
        res = pd.concat([old_data, new_data])
    else:
        res = new_data
    if old_chain is not None:
        chain = old_chain + chain
    chain = chain - _data
    return dictattr(data = res, chain = chain, roll = roll)

df_roll_off.output = ['data', 'roll', 'chain']


import pandas as pd; import numpy as np
from pyg_base._inspect import getargs, getcallarg, getcallargs, getargspec, call_with_callargs
from pyg_base._types import is_ts, is_str, is_df, is_pd, is_series, is_arr, is_array, is_tuple, is_dict
from pyg_base._decorators import wrapper
from pyg_base._as_list import as_tuple, as_list

__all__ = ['loop', 'loops', 'len0', 'pd2np', 'shape', 'loop_all']


def _zero():
    return 0

def len0(value):
    """
    returns the len of an object or 0 if no len(value) exists
    
    :Example:
    ---------
    >>> assert len0(5) == 0
    >>> assert len0('a string is dimensionless') == 0
    >>> assert len0([]) == 0
    >>> assert len0([1,2]) == 2

    Parameters
    ----------
    value : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    try:
        return 0 if is_str(value) else getattr(value, '__len__', _zero)()
    except Exception:
        return 0


def shape(value):
    return getattr(value, 'shape', ())
 

def _item_by_key(value, key, keys, i = None):
    if isinstance(value, dict):
        if sorted(value.keys()) == keys:
            return value[key]
        else:
            return type(value)({k : _item_by_key(v, key, keys, i) for k, v in value.items()})
    elif isinstance(value, pd.Series) and sorted(value.index.values) == keys:
        return value[key] 
    elif isinstance(value, pd.DataFrame) and sorted(value.columns) == sorted(keys):
        return value[key]
    elif isinstance(value, pd.DataFrame) and sorted(value.index) == sorted(keys):
        return value.loc[key]
    elif is_array(value) and len(value.shape):
        if len(value.shape) == 2 and value.shape[1] == len(keys) and i is not None:
            return value.T[i]
        elif len(value.shape) == 1 and value.shape[0] == len(keys) and i is not None:
            return value[i]
        else:
            return value
    else:
        if i is not None and len0(value) == len(keys):
            return value[i]
        else:
            return value

def _item_by_i(value, i, n):
    if is_df(value) and value.shape[1] == 1:
        value = value.iloc[:,0]
    elif is_array(value) and len(value.shape) == 2 and value.shape[1] == 1:
        value = value.T[0]
    if isinstance(value, (list, tuple)):
        if len(value) == n:
            return value[i]
        else:
            return type(value)([_item_by_i(v, i, n) for v in value])
    elif is_array(value):
        if len(value.shape) == 2 and value.shape[-1] == n:
            return value.T[i]
        if len(value.shape) == 1 and value.shape[-1] == n:
            return value[i]
    elif is_series(value) and len(value) == n and not is_ts(value):
        return value.iloc[i]
    elif is_df(value) and value.shape[1] == n:
        return value.iloc[:, i]
    return value           

# add_index_and_columns = _np2pd

def add_index_and_columns(res, arg):
    if isinstance(res, tuple):
        return tuple([add_index_and_columns(r, arg) for r in res])
    elif isinstance(res, dict):
        return type(res)({k: add_index_and_columns(r, arg) for k,r in res.items()})
    if isinstance(res, (pd.Series, pd.DataFrame)) and len(res) == len(arg):
        res.index = arg.index
    if isinstance(res, (pd.DataFrame)) and res.shape[1] == arg.shape[1]:
        res.columns = arg.columns
    return res

def axis0_to_dataframe(res, arg):
    if min([is_tuple(r) for r in res], default = False) and len(set([len(r) for r in res])) == 1:
        return tuple([axis0_to_dataframe(r, arg) for r in list(zip(*res))])
    elif min([is_dict(r) for r in res], default = False) and len(set([len(r) for r in res])) == 1:
        keys = res[0].keys()
        tp = type(res[0])
        return tp({key : axis0_to_dataframe([r[key] for r in res], arg) for key in keys})
    try:
        res = type(arg)(dict(zip(arg.columns, res)))
        return add_index_and_columns(res, arg)
    except ValueError:
        res = pd.Series(res)
        res.index = arg.columns
        return res

def axis0_to_array(res, arg):
    if min([is_tuple(r) for r in res], default = False) and len(set([len(r) for r in res])) == 1:
        return tuple([axis0_to_array(r, arg) for r in list(zip(*res))])
    elif min([is_dict(r) for r in res], default = False) and len(set([len(r) for r in res])) == 1:
        keys = res[0].keys()
        tp = type(res[0])
        return tp({key : axis0_to_array([r[key] for r in res], arg) for key in keys})
    rtn = np.array(res)
    if len(rtn.shape)>1:
        rtn = rtn.T
    return rtn

class loops(wrapper):

    """
    
    converts a function to loop over the arguments, depending on the type of the first argument

    :Examples:
    --------------
    >>> @loop(dict, list, pd.DataFrame, pd.Series)
    >>> def f(a,b):
    >>>     return a+b
    
    >>> assert f(1,2) == 3
    >>> assert f([1,2,3],2) == [3,4,5]
    >>> assert f([1,2,3], [4,5,6]) == [5,7,9]    
    
    >>> assert f(dict(x=1,y=2), 3) == dict(x = 4, y = 5)
    >>> assert f(dict(x=1,y=2), dict(x = 3, y = 4)) == dict(x = 4, y = 6)
    

    >>> a = pd.Series(dict(x=1,y=2))
    >>> b = dict(x=3,y=4)
    >>> assert np.all(f(a,b) == pd.Series(dict(x=4,y=6)))    

    >>> a = pd.DataFrame(dict(x=[1,1],y=[2,2])); a.index = [5,10]
    >>> b = dict(x=3,y=4)
    >>> res =  f(a,b)
    >>> assert np.all(res == pd.DataFrame(dict(x=[4,4],y=[6,6]), index = [5,10]))    
    
    >>> a = pd.DataFrame(dict(x=[1,1],y=[2,2])); a.index = [5,10]
    >>> res =  f(a,[3,4])
    >>> assert np.all( res == pd.DataFrame(dict(x=[4,4],y=[6,6]), index = [5,10]))   
    
    >>> from pyg import * 
    >>> @loop_all
    >>> def f(x, y, col):
    >>>     return x * y[col]
    >>> x = dict(a = 1, b = 2)
    >>> y = pd.DataFrame(dict(col1 = [1,2], col2 = [3,4]), index = ['a', 'b'])
    >>> assert f(x, y, 'col1') == dict(a = 1, b = 4)
    >>> assert f(x, y, 'col2') == dict(a = 3, b = 8)
    
    """

    def __init__(self, function = None, types = None):
        super(loops, self).__init__(function = function, types = as_tuple(types))

    @property
    def first(self):
        args = getargs(self)
        return args[0] if len(args) else None
    
    def T(self, arg, args, kwargs):
        arg_ = _T(arg)
        args_ = _T(args)
        kwargs_ = _T(kwargs)
        res = self._wrapped(arg_, args_, kwargs_)
        return _T(res)

    def wrapped(self, *args, **kwargs):
        top = self.first
        if len(args) == 0 and not top in kwargs:
            return self.function(*args, **kwargs)
        if len(args):
            arg, args_, kwargs_ = args[0], args[1:], kwargs
        else:
            arg = kwargs.pop(top)
            args_, kwargs_ = args, kwargs
        if isinstance(arg, pd.Series) and pd.Series in self.types and not is_ts(arg):
            keys = sorted(arg.index)
            res = {key : self._wrapped(arg[key], (_item_by_key(a,key,keys) for a in args_), {k : _item_by_key(v,key,keys) for k,v in kwargs_.items()}) for key in arg.index}
            return type(arg)(res)          
        else:
            return self._wrapped(arg, args_, kwargs_)
    
    def _wrapped(self, arg, args, kwargs):
        axis = kwargs.pop('axis', 0)
        if isinstance(arg, dict) and dict in self.types:
            keys = sorted(arg.keys())
            res = {key : self._wrapped(arg[key], (_item_by_key(a,key,keys) for a in args), {k : _item_by_key(v,key,keys) for k,v in kwargs.items()}) for key in arg.keys()}
            return type(arg)(res)
        elif isinstance(arg, pd.DataFrame) and pd.DataFrame in self.types:
            if axis in (1,-1):
                res = self.T(arg, args, kwargs)
                if isinstance(res, (pd.DataFrame, pd.Series)) and len(res) == len(arg):
                    res.index = arg.index
                return add_index_and_columns(res, arg)
            else:
                keys = sorted(arg.columns)
                res = [self._wrapped(arg[key], (_item_by_key(a,key,keys,i) for a in args), {k : _item_by_key(v,key,keys,i) for k,v in kwargs.items()}) for i, key in enumerate(arg.columns)]
                rtn = axis0_to_dataframe(res, arg)
                return rtn
        elif isinstance(arg, pd.Series):
            return self.function(arg, *args, **kwargs)
        elif isinstance(arg, np.ndarray) and np.ndarray in self.types:
            if len(arg.shape) <= 1:
                return self.function(arg, *args, **kwargs) 
            else:
                if axis in (1,-1):
                    return self.T(arg, args, kwargs)
                else:
                    n = arg.shape[1]
                    res = [self._wrapped(_item_by_i(arg,i,n), (_item_by_i(a,i,n) for a in args), {k: _item_by_i(v,i,n) for k, v in kwargs.items()}) for i in range(n)]
                    return axis0_to_array(res, arg)
        elif isinstance(arg, self.types):
            n = len(arg)
            res = [self._wrapped(arg[i], (_item_by_i(a,i,n) for a in args), {k: _item_by_i(v,i,n) for k, v in kwargs.items()}) for i in range(n)]                            
            return type(arg)(res)
        else:
            return self.function(arg, *args, **kwargs)

def loop(*types):
    """
    returns an instance of loops(types = types)
    """
    return loops(types = as_tuple(types))


def _T(arg):
    if isinstance(arg, tuple):
        return type(arg)([_T(a) for a in arg])
    elif isinstance(arg, list):
        return type(arg)([_T(a) for a in arg])
    elif isinstance(arg, dict):
        return type(arg)({k :_T(a) for k,a in arg.items()})
    else:
        return arg.T if isinstance(arg, (np.ndarray, pd.DataFrame)) and len(arg.shape) == 2 else arg

def _values(a):
    if isinstance(a, (list, tuple)):
        return type(a)([_values(v) for v in a])
    elif isinstance(a, dict):
        return type(a)({k : _values(v) for k,v in a.items()})
    if is_series(a):
        return a.values
    elif is_df(a):
        if a.shape[-1] == 1:
            return a.values.T[0]
        else:
            return a.values
    else:
        return a

def _np2pd(res, arg):
    if isinstance(res, tuple):
        return tuple([_np2pd(r, arg) for r in res])
    elif isinstance(res, tuple):
        return [_np2pd(r, arg) for r in res]
    elif isinstance(res, dict):
        return type(res)({k: _np2pd(r, arg) for k,r in res.items()})
    n = len0(res)
    if n == len0(arg):
        if len(res.shape) == 2:
            res = pd.DataFrame(res, arg.index)
            if len(arg.shape) == 2 and res.shape[1] == arg.shape[1]:
                res.columns = arg.columns
            return res
        else:
            return pd.Series(res, arg.index)
    elif is_df(arg) and n == arg.shape[1]:
        return pd.Series(arg.columns, res)
    else:
        return res

class pd2np(wrapper):
    """
    converts a numpy-based function to work on pandas by converting to a numpy arrage and then converting back to panadas dataframe
    """
    def wrapped(self, *args, **kwargs):
        arg = getcallarg(self.function, args, kwargs)
        if not is_pd(arg):
            return self.function(*args, **kwargs)
        args_ = [_values(a) for a in args]
        kwargs_ = {k : _values(v) for k, v in kwargs.items()}
        res = self.function(*args_, **kwargs_)
        return _np2pd(res, arg)


loop_all = loops(types = (pd.Series, pd.DataFrame, np.ndarray, list, tuple, dict))


@loop(dict, list, tuple)
def _cut_pd(arg, index):
    if is_pd(arg):
        return arg.drop(index = index)
    elif isinstance(arg, pd.Index):
        return arg.drop(index) # index just has labels
    else:
        return arg

@loop(dict, list, tuple)
def _cut_np(arg, n, t = None):
    if is_arr(arg) and (t is None or len(arg) == t):
        return arg[n:]
    else:
        return arg



class skip_if_data_pd(wrapper):
    """
    If a function has a 'data' parameter which is provided and is a timeseries, use it to start running from that point onwards
    
    :Example:
    ---------
    >>> a = pd.Series(range(6))
    >>> f = skip_if_data_pd(lambda a: a ** 2)    
    >>> assert list(f(a).values) == [0,1,4,9,16,25]
    
    >>> res = f(a, data = a.iloc[:4]) ## data is provided and is assumed to be the history of past calculations, regardless if it is actual value
    >>> assert list(res.values) == [0,1,2,3] + [16,25]
    
    """
    def wrapped(self, *args, **kwargs):
        data = kwargs.pop('data', None)
        if is_pd(data):
            data = data.iloc[:-1]
            index = data.index
            args, kwargs = _cut_pd((args, kwargs), index = index)
        res = self.function(*args, **kwargs)
        if res is None: ## error handling
            return None
        if is_pd(data):
            if not is_pd(res):
                raise ValueError('if data parameter %s is a dataframe, it is part of the answer so result must also be a dataframe %s'%(data, res))                    
            return pd.concat([data,res]).sort_index()
        else:
            return res

def _np2pd_(res, index = None, columns = None):
    if is_pd(res) or not is_arr(res) or len(res.shape)>2 or (columns is None and index is None):
        return res
    elif len(res.shape) == 1:
        if index is not None and len(res) == len(index):
            return pd.Series(res, index)
        if columns is not None and len(res) == len(columns):
            return pd.Series(res, columns)
        else:
            return res
    elif len(res.shape) == 2:
        return pd.DataFrame(data = res, columns = columns, index = index) 
    else:
        return res

class skip_if_data_pd_or_np(wrapper):
    """
    If a function has a 'data' parameter which is provided and is a timeseries or a numpy array, use it to start running from that point onwards
    
    :Example:
    ---------
    >>> a = np.array([0,1,2,3,4,5])
    >>> f = skip_if_data_pd_or_np(lambda a: a ** 2)    
    >>> assert list(f(a)) == [0,1,4,9,16,25]
    
    >>> res = f(a, data = a[:4]) ## data is provided and is assumed to be the history of past calculations, regardless if it is actual value
    >>> assert list(res) == [0,1,2,3] + [16,25]
    
    """
    def wrapped(self, *args, **kwargs):
        data = kwargs.pop('data', None)
        arg = getcallarg(self.function, args, kwargs)
        index = kwargs.pop('index', arg.index if is_pd(arg) else None)
        columns = kwargs.pop('columns', arg.columns if is_df(arg) else None)
        if index is not None:
            index = pd.Index(index)
        if data is not None:
            t = None
            if index is not None:
                t = len(index)
            else:
                if is_arr(arg) or is_pd(arg):
                    t = len(arg)
            if is_pd(data): ## previous result is a timeseries
                args, kwargs, index = _cut_pd((args, kwargs, index), index = data.index)
            args, kwargs, index = _cut_np((args, kwargs, index), n = len(data), t = t)
        res = self.function(*args, **kwargs)        
        if res is None: ## error handling
            return None
        if is_arr(data):
            if not is_arr(res):
                raise ValueError('if data parameter %s is a numpy array, it is part of the answer so result must also be a dataframe %s'%(data, res))                                    
            return np.concatenate([data,res])            
        res_df = _np2pd_(res, index, columns)
        if is_pd(data):
            if not is_pd(res_df):
                raise ValueError('if data parameter %s is a dataframe, it is part of the answer so result must also be a dataframe %s'%(data, res_df))                                    
            return pd.concat([data,res]).sort_index()
        return res_df

def _getitem(value, param, strict = False):
    if isinstance(value, dict) and param in value.keys():
        return value[param]
    elif isinstance(value, pd.Series) and param in value.index:
        return value[param]
    elif isinstance(value, pd.DataFrame) and param in value.columns:
        return value[param]
    elif isinstance(value, pd.DataFrame) and param in value.index:
        return value.loc[param]
    else:
        if strict and isinstance(value, (pd.Series, pd.DataFrame, dict)):
            raise ValueError('could not find parameter %s in %s'%(param, value))
        else:
            return value

class grab_parameter_from_dict(wrapper):
    """
    allows the function to provide a "data factory" for a function.
    
    
    Example
    -------
    >>> @grab_parameter_from_dict
    >>> def f(a,b):
    >>>     return a+b
    
    >>> data_factory = dict(a = 1, b = 2)
    >>> assert f(a = data_factory, b = data_factory) == 3
    
    >>> data_factory = pd.Series(dict(a = 10, b = 20))
    >>> assert f(a = data_factory, b = data_factory) == 30
    
    :Example: being specific about what parameters are converted
    ---------    
    
    >>> @grab_parameter_from_dict(parameters = 'a') # convert just 'a'
    >>> def f(a,b):
    >>>     return a+b

    >>> data_factory = pd.Series(dict(a = 10, b = 20))
    >>> assert eq(f(a = data_factory, b = data_factory), data_factory['a'] + data_factory)
    
    ### From example, we see that there may be ambiguity on using the factory and conversion...
    
    :Example: being strict about which parameters are converted
    --------
    >>> @grab_parameter_from_dict(parameters = 'a', strict = False) # convert just 'a'
    >>> def f(a,b):
    >>>     return a+b
    
    >>> bad_data_factory = pd.Series(dict(c = 20))
    >>> f(a = bad_data_factory, b = 5) = bad_data_factory + 5
    >>> ## This is a problem, as 'a' should have converted to a value
    
    >>> @grab_parameter_from_dict(parameters = 'a', strict = True) # convert just 'a', be strict
    >>> def f(a,b):
    >>>     return a+b
    
    >>> import pytest
    >>> with pytest.raises(ValueError):
    >>>     f(a = bad_data_factory, b = 5)

    >>> assert f(a = 3, b = 5) == 8
    >>> #This is still OK as 'a' is not a type dict/pd.Series or pd.DataFrame

    :Example: remapping the parameters
    ---------
    >>> @grab_parameter_from_dict(a = 'fancy_name') # convert just 'a' using 'fancy_name' in data factory
    >>> def f(a,b):
    >>>     return a+b
    >>>
    >>> factory = dict(fancy_name = 4)
    >>> assert f(a = factory, b = 3) == 7
    
    """
    
    def __init__(self, function = None, parameters = None, strict = None, **kwargs):
        if strict is None:
            strict = False if parameters is None else True
        if isinstance(parameters, (list, str)):
            parameters = as_list(parameters)
            parameters = dict(zip(parameters, parameters))
        if kwargs:
            parameters = parameters or {}
            parameters.update(kwargs)
        super(grab_parameter_from_dict, self).__init__(function = function, parameters = parameters, strict = strict)

    def wrapped(self, *args, **kwargs):
        spec = getargspec(self.function)
        callargs = getcallargs(self.function, *args, **kwargs)
        strict = self.strict
        if self.parameters is None:
            callargs = {param: value if param == spec.varargs else _getitem(value, param, strict = strict) for param, value in callargs.items()}
            result = call_with_callargs(self.function, callargs)
        else:
            parameters = self.parameters
            callargs = {param: _getitem(value, parameters[param], strict = strict) if param in parameters else value for param, value in callargs.items()}
            result = call_with_callargs(self.function, callargs)
        return result
